"""Generate lists of most popular collections for each of the hardcoded DAACs."""

from pathlib import Path

import requests
from earthaccess.daac import DAACS

THIS_DIR = Path(__file__).parent

# TODO: Can we query CMR for all providers, instead of relying on hard-coded names in a DAAC list?
#   For example, using this URL: "https://cmr.earthdata.nasa.gov/ingest/providers"
all_providers = [
    provider
    for daac_info in DAACS
    for provider in set(
        list(daac_info["cloud-providers"]) + list(daac_info["on-prem-providers"])
    )
]


def top_collections(
    *,
    provider: str,
    num: int = 100,
) -> list[str]:
    if num > 2000:
        raise RuntimeError(
            "Paging not supported, can only get up to 2000 top collections"
        )

    response = requests.post(
        "https://cmr.earthdata.nasa.gov/search/collections.json",
        data={
            "provider": provider,
            "has_granules_or_cwic": True,
            "include_facets": "v2",
            "include_granule_counts": True,
            "include_has_granules": True,
            "include_tags": "edsc.*,opensearch.granule.osdd",
            "page_num": 1,
            "page_size": num,
            "sort_key[]": "-usage_score",
        },
    )
    collection_ids = [
        collection["id"] for collection in response.json()["feed"]["entry"]
    ]
    return collection_ids


def main():
    for provider in all_providers:
        collection_ids = top_collections(provider=provider)

        output = THIS_DIR / f"{provider}.txt"
        output.write_text("\n".join(collection_ids) + "\n")


if __name__ == "__main__":
    main()



import fsspec
import xarray as xr
from earthaccess.store import EarthAccessFile


def test_open_dataset():
    ds = xr.Dataset({"y": ("x", list(range(8)))})
    fs = fsspec.filesystem("memory")
    path = "foo"
    with fs.open(path, mode="wb") as f:
        ds.to_netcdf(f, engine="h5netcdf")
    f = fs.open(path)
    earthaccess_file = EarthAccessFile(f, granule="foo")
    # confirm backend detection
    backends = xr.backends.list_engines()
    assert backends["h5netcdf"].guess_can_open(earthaccess_file)
    # confirm opens correctly
    assert (xr.open_dataset(earthaccess_file) == ds).all()
    # cleanup
    fs.store.clear()



from __future__ import annotations

import getpass
import importlib.metadata
import logging
import os
import platform
import shutil
from netrc import NetrcParseError
from pathlib import Path
from typing import Any, Dict, Mapping, Optional
from urllib.parse import urlparse

import requests
import requests.cookies
from tinynetrc import Netrc
from typing_extensions import deprecated

from earthaccess.daac import DAACS
from earthaccess.exceptions import LoginAttemptFailure, LoginStrategyUnavailable
from earthaccess.system import PROD, System

try:
    user_agent = f"earthaccess v{importlib.metadata.version('earthaccess')}"
except importlib.metadata.PackageNotFoundError:
    user_agent = "earthaccess"


logger = logging.getLogger(__name__)


def netrc_path() -> Path:
    """Return the path of the `.netrc` file.

    The path may or may not exist.

    See [the `.netrc` file](https://www.gnu.org/software/inetutils/manual/html_node/The-_002enetrc-file.html).

    Returns:
        `Path` of the `NETRC` environment variable, if the value is non-empty;
        otherwise, the path of the platform-specific default location:
        `~/_netrc` on Windows systems, `~/.netrc` on non-Windows systems.
    """
    sys_netrc_name = "_netrc" if platform.system() == "Windows" else ".netrc"
    env_netrc = os.environ.get("NETRC")

    return Path(env_netrc) if env_netrc else Path.home() / sys_netrc_name


class BasicAuthResponseHook:
    def __init__(self, hostname: str, auth: tuple[str, str]) -> None:
        self.hostname = hostname
        self.auth = auth

    def __call__(self, r: requests.Response, **kwargs: Any) -> requests.Response:
        from http.cookiejar import CookieJar

        # If the response's URL is not for the EDL system we're authenticating
        # against, then simply return the response unchanged.  Otherwise, we'll
        # prepare a new request below with the user's EDL credentials.
        if urlparse(r.url).hostname != self.hostname:
            return r

        # Consume content and release the original connection to allow our new
        # request to reuse the same one.
        r.content
        r.close()

        prepared_request = r.request.copy()
        cookies: CookieJar = prepared_request._cookies  # type: ignore
        requests.cookies.extract_cookies_to_jar(cookies, r.request, r.raw)
        prepared_request.prepare_cookies(cookies)
        prepared_request.prepare_auth(self.auth)

        new_r = r.connection.send(prepared_request, **kwargs)
        new_r.history.append(r)
        new_r.request = prepared_request

        return new_r


class SessionWithHeaderRedirection(requests.Session):
    """Requests removes auth headers if the redirect happens outside the
    original req domain.
    """

    def __init__(self, edl_hostname: str, auth: tuple[str, str] | None = None) -> None:
        super().__init__()
        self.headers.update({"User-Agent": user_agent})

        if auth:
            hook = BasicAuthResponseHook(edl_hostname, auth)
            self.hooks["response"].append(hook)


class Auth(object):
    """Authentication class for operations that require Earthdata login (EDL)."""

    def __init__(self) -> None:
        # Maybe all these predefined URLs should be in a constants.py file
        self.authenticated = False
        self.token: Mapping[str, str] | None = None
        self.username: str | None = None
        self.password: str | None = None
        self._set_earthdata_system(PROD)

    def login(
        self,
        strategy: str = "netrc",
        persist: bool = False,
        system: Optional[System] = None,
    ) -> Any:
        """Authenticate with Earthdata login.

        Parameters:
            strategy:
                The authentication method.

                * **"interactive"**: Enter a username and password.
                * **"netrc"**: (default) Retrieve a username and password from ~/.netrc.
                * **"environment"**:
                    Retrieve either a username and password pair from the
                    `EARTHDATA_USERNAME` and `EARTHDATA_PASSWORD` environment variables,
                    or an Earthdata login token from the `EARTHDATA_TOKEN` environment
                    variable.
            persist: Will persist username and password credentials in a `.netrc` file.
            system: the EDL endpoint to log in to Earthdata, defaults to PROD

        Returns:
            An instance of Auth.

        Raises:
            LoginAttemptFailure: If the NASA Earthdata Login service rejects
                credentials.
        """
        if system is not None:
            self._set_earthdata_system(system)

        if self.authenticated and (system == self.system):
            logger.debug("We are already authenticated with NASA EDL")
            return self

        if strategy == "interactive":
            self._interactive(persist)
        elif strategy == "netrc":
            self._netrc()
        elif strategy == "environment":
            self._environment()

        return self

    def _set_earthdata_system(self, system: System) -> None:
        self.system = system

        # Maybe all these predefined URLs should be in a constants.py file
        self.EDL_FIND_OR_CREATE_TOKEN_URL = (
            f"https://{self.system.edl_hostname}/api/users/find_or_create_token"
        )

        self._eula_url = (
            f"https://{self.system.edl_hostname}/users/earthaccess/unaccepted_eulas"
        )
        self._apps_url = f"https://{self.system.edl_hostname}/application_search"

    @deprecated("No replacement, as tokens are now refreshed automatically.")
    def refresh_tokens(self) -> bool:
        """Refresh CMR tokens.

        Tokens are used to do authenticated queries on CMR for restricted and early access datasets.
        This method renews the tokens to make sure we can query the collections allowed to our EDL user.
        """
        return self.authenticated

    def get_s3_credentials(
        self,
        daac: Optional[str] = None,
        provider: Optional[str] = None,
        endpoint: Optional[str] = None,
    ) -> Dict[str, str]:
        """Gets AWS S3 credentials for a given NASA cloud provider.

        The easier way is to use the DAAC short name; provider is optional if we know it.

        Parameters:
            daac: The name of a NASA DAAC, e.g. NSIDC or PODAAC.
            provider: A valid cloud provider. Each DAAC has a provider code for their cloud distributions.
            endpoint: Getting the credentials directly from the S3Credentials URL.

        Returns:
            A Python dictionary with the temporary AWS S3 credentials.
        """
        if not self.authenticated:
            logger.info("We need to authenticate with EDL first")
            return {}

        auth_url = endpoint or self._get_cloud_auth_url(
            daac_shortname=daac, provider=provider
        )

        if not auth_url.startswith("https://"):
            # This happens if the cloud provider doesn't list the S3 credentials or the DAAC
            # does not have cloud collections yet
            logger.info(f"Credentials for the cloud provider {daac} are not available")
            return {}

        with self.get_session() as session, session.get(auth_url, timeout=15) as r:
            if r:
                return r.json()

            logger.error(
                f"Authentication with Earthdata Login failed with:\n{r.text[:1000]}"
            )
            logger.error(
                f"Consider accepting the EULAs available at {self._eula_url} and applications at {self._apps_url}"
            )

            return {}

    def get_session(self) -> requests.Session:
        """Returns a new request session instance.

        Parameters:
            bearer_token: whether to include bearer token

        Returns:
            class Session instance with Auth and bearer token headers
        """
        username, password = self.username, self.password
        auth = (username, password) if username and password else None
        session = SessionWithHeaderRedirection(self.system.edl_hostname, auth)

        if self.token:
            session.headers["Authorization"] = f"Bearer {self.token['access_token']}"

        return session

    def _interactive(
        self,
        persist_credentials: bool = False,
    ) -> bool:
        username = input("Enter your Earthdata Login username: ")
        password = getpass.getpass(prompt="Enter your Earthdata password: ")
        authenticated = self._get_credentials(username, password, None)
        if authenticated:
            logger.debug("Using user provided credentials for EDL")
            if persist_credentials:
                self._persist_user_credentials(username, password)
        return authenticated

    def _netrc(self) -> bool:
        netrc_loc = netrc_path()

        try:
            my_netrc = Netrc(str(netrc_loc))
        except FileNotFoundError as err:
            raise LoginStrategyUnavailable(f"No .netrc found at {netrc_loc}") from err
        except NetrcParseError as err:
            raise LoginStrategyUnavailable(
                f"Unable to parse .netrc file {netrc_loc}"
            ) from err

        creds = my_netrc[self.system.edl_hostname]
        if creds is None:
            raise LoginStrategyUnavailable(
                f"Earthdata Login hostname {self.system.edl_hostname} not found in .netrc file {netrc_loc}"
            )

        username = creds["login"]
        password = creds["password"]

        if username is None:
            raise LoginStrategyUnavailable(
                f"Username not found in .netrc file {netrc_loc}"
            )
        if password is None:
            raise LoginStrategyUnavailable(
                f"Password not found in .netrc file {netrc_loc}"
            )

        authenticated = self._get_credentials(username, password, None)

        if authenticated:
            logger.debug("Using .netrc file for EDL")

        return authenticated

    def _environment(self) -> bool:
        username = os.getenv("EARTHDATA_USERNAME")
        password = os.getenv("EARTHDATA_PASSWORD")
        token = os.getenv("EARTHDATA_TOKEN")

        if (not username or not password) and not token:
            raise LoginStrategyUnavailable(
                "Either the environment variables EARTHDATA_USERNAME and "
                "EARTHDATA_PASSWORD must both be set, or EARTHDATA_TOKEN must be set for "
                "the 'environment' login strategy."
            )

        logger.debug("Using environment variables for EDL")
        return self._get_credentials(username, password, token)

    def _get_credentials(
        self,
        username: Optional[str],
        password: Optional[str],
        user_token: Optional[str],
    ) -> bool:
        if user_token is not None:
            self.token = {"access_token": user_token}
            self.authenticated = True
        elif username is not None and password is not None:
            self.username = username
            self.password = password
            token_resp = self._find_or_create_token()

            if not (token_resp.ok):  # type: ignore
                msg = f"Authentication with Earthdata Login failed with:\n{token_resp.text}"
                logger.error(msg)
                raise LoginAttemptFailure(msg)

            logger.info("You're now authenticated with NASA Earthdata Login")

            token = token_resp.json()
            logger.info("Using token with expiration date %s", token["expiration_date"])
            self.token = token
            self.authenticated = True

        return self.authenticated

    def _find_or_create_token(self) -> requests.Response:
        with self.get_session() as session:
            return session.post(
                self.EDL_FIND_OR_CREATE_TOKEN_URL,
                headers={"Accept": "application/json"},
                timeout=10,
            )

    def _persist_user_credentials(self, username: str, password: str) -> bool:
        # See: https://github.com/sloria/tinynetrc/issues/34

        netrc_loc = netrc_path()
        logger.info(f"Persisting credentials to {netrc_loc}")

        try:
            netrc_loc.touch(exist_ok=True)
            netrc_loc.chmod(0o600)
        except Exception as e:
            logger.error(e)
            return False

        my_netrc = Netrc(str(netrc_loc))
        my_netrc[self.system.edl_hostname] = {
            "login": username,
            "password": password,
        }
        my_netrc.save()

        urs_cookies_path = Path.home() / ".urs_cookies"

        if not urs_cookies_path.exists():
            urs_cookies_path.write_text("")

        # Create and write to .dodsrc file
        dodsrc_path = Path.home() / ".dodsrc"

        if not dodsrc_path.exists():
            dodsrc_contents = (
                f"HTTP.COOKIEJAR={urs_cookies_path}\nHTTP.NETRC={netrc_loc}"
            )
            dodsrc_path.write_text(dodsrc_contents)

        if platform.system() == "Windows":
            local_dodsrc_path = Path.cwd() / dodsrc_path.name

            if not local_dodsrc_path.exists():
                shutil.copy2(dodsrc_path, local_dodsrc_path)

        return True

    def _get_cloud_auth_url(
        self, daac_shortname: Optional[str] = "", provider: Optional[str] = ""
    ) -> str:
        for daac in DAACS:
            if (
                daac_shortname == daac["short-name"]
                or provider in daac["cloud-providers"]
                and len(daac["s3-credentials"]) > 0
            ):
                return str(daac["s3-credentials"])
        return ""



import datetime
import logging
import threading
import traceback
from functools import lru_cache
from itertools import chain
from pathlib import Path
from pickle import dumps, loads
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
from uuid import uuid4

import fsspec
import requests
import s3fs
from multimethod import multimethod as singledispatchmethod
from pqdm.threads import pqdm
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from typing_extensions import deprecated

import earthaccess

from .auth import Auth
from .daac import DAAC_TEST_URLS, find_provider
from .exceptions import DownloadFailure, EulaNotAccepted
from .results import DataGranule
from .search import DataCollections

logger = logging.getLogger(__name__)


def _is_interactive() -> bool:
    """Detect if earthaccess is being used in an interactive session.
    Interactive sessions include Jupyter Notebooks, IPython REPL, and default Python REPL.
    """
    try:
        from IPython import get_ipython  # type: ignore

        # IPython Notebook or REPL:
        if get_ipython() is not None:
            return True
    except ImportError:
        pass

    import sys

    # Python REPL
    return hasattr(sys, "ps1")


class EarthAccessFile:
    """Handle for a file-like object pointing to an on-prem or Earthdata Cloud granule."""

    def __init__(
        self, f: fsspec.spec.AbstractBufferedFile, granule: DataGranule
    ) -> None:
        """EarthAccessFile connects an Earthdata search result with an open file-like object.

        The class implements custom serialization, but otherwise passes all attribute and method calls
        directly to the file-like object given during initialization. An instance of
        this class can be treated like that file-like object itself.

        Note that `type()` applied to an instance of this class is expected to disagree with
        the `__class__` attribute on the instance.

        Parameters:
            f: a file-like object
            granule: a granule search result
        """
        self.f = f
        self.granule = granule

    def __getattribute__(self, name: str) -> Any:
        # use super().__getattribute__ to avoid infinite recursion
        if (name in EarthAccessFile.__dict__) or (name in self.__dict__):
            # accessing our attributes
            return super().__getattribute__(name)
        else:
            # access proxied attributes
            proxy = super().__getattribute__("f")
            return getattr(proxy, name)

    def __reduce_ex__(self, protocol: Any) -> Any:
        return make_instance, (
            self.__class__,
            self.granule,
            earthaccess.__auth__,
            dumps(self.f),
        )

    def __repr__(self) -> str:
        return repr(self.f)


def _optimal_fsspec_block_size(file_size: int) -> int:
    """Determine the optimal block size based on file size.
    Note: we could even be smarter if we know the chunk sizes of the variables
    we need to cache, e.g. using the `dmrpp` file and the `wellknownparts` cache type.

    Uses `blockcache` for all files with block sizes adjusted by file size:

    - <100MB: 4MB
    - >100MB: 4–16MB

    Parameters:
        file_size (int): Size of the file in bytes.

    Returns:
        block_size (int): Optimal block size in bytes.
    """
    if file_size < 100 * 1024 * 1024:
        block_size = 4 * 1024 * 1024
    elif 100 * 1024 * 1024 <= file_size < 1024 * 1024 * 1024:
        block_size = 8 * 1024 * 1024
    else:
        block_size = 16 * 1024 * 1024

    return block_size


def _open_files(
    url_mapping: Mapping[str, Union[DataGranule, None]],
    fs: fsspec.AbstractFileSystem,
    *,
    pqdm_kwargs: Optional[Mapping[str, Any]] = None,
    open_kwargs: Optional[Dict[str, Any]] = None,
) -> List[fsspec.spec.AbstractBufferedFile]:
    def multi_thread_open(data: tuple[str, Optional[DataGranule]]) -> EarthAccessFile:
        url, granule = data
        f_size = fs.info(url)["size"]
        default_cache_type = "background"  # block cache with background fetching
        default_block_size = _optimal_fsspec_block_size(f_size)

        open_kw = (open_kwargs or fsspec.config.conf or {}).copy()

        open_kw.setdefault("cache_type", default_cache_type)
        open_kw.setdefault("block_size", default_block_size)

        f = fs.open(url, **open_kw)
        return EarthAccessFile(f, granule)  # type: ignore

    # this {#n_jobs} is for the unittests as this method is not public and pqdm will have values at this point
    return pqdm(
        url_mapping.items(), multi_thread_open, **(pqdm_kwargs or {"n_jobs": 8})
    )


def make_instance(
    cls: Any, granule: DataGranule, auth: Auth, data: Any
) -> EarthAccessFile:
    # Attempt to re-authenticate
    if not earthaccess.__auth__.authenticated:
        earthaccess.__auth__ = auth
        earthaccess.login()

    # When sending EarthAccessFiles between processes, it's possible that
    # we will need to switch between s3 <--> https protocols.
    if (earthaccess.__store__.in_region and cls is not s3fs.S3File) or (
        not earthaccess.__store__.in_region and cls is s3fs.S3File
    ):
        # NOTE: This uses the first data_link listed in the granule. That's not
        #       guaranteed to be the right one.
        return earthaccess.open([granule])[0]
    else:
        return EarthAccessFile(loads(data), granule)


def _get_url_granule_mapping(
    granules: List[DataGranule], access: str
) -> Mapping[str, DataGranule]:
    """Construct a mapping between file urls and granules."""
    url_mapping = {}
    for granule in granules:
        for url in granule.data_links(access=access):
            url_mapping[url] = granule
    return url_mapping


class Store(object):
    """Store class to access granules on-prem or in the cloud."""

    def __init__(self, auth: Any, pre_authorize: bool = False) -> None:
        """Store is the class to access data.

        Parameters:
            auth: Auth instance to download and access data.
        """
        self.thread_locals = threading.local()
        if auth.authenticated is True:
            self.auth = auth
            self._s3_credentials: Dict[
                Tuple, Tuple[datetime.datetime, Dict[str, str]]
            ] = {}
            oauth_profile = f"https://{auth.system.edl_hostname}/profile"
            # sets the initial URS cookie
            self._requests_cookies: Dict[str, Any] = {}
            self.set_requests_session(oauth_profile)
            if pre_authorize:
                # collect cookies from other DAACs
                for url in DAAC_TEST_URLS:
                    self.set_requests_session(url)

        else:
            logger.warning("The current session is not authenticated with NASA")
            self.auth = None
        self.in_region = self._running_in_us_west_2()

    def _derive_concept_provider(self, concept_id: Optional[str] = None) -> str:
        if concept_id is not None:
            provider = concept_id.split("-")[1]
            return provider
        return ""

    def _derive_daac_provider(self, daac: str) -> Union[str, None]:
        provider = find_provider(daac, True)
        return provider

    def _is_cloud_collection(self, concept_id: List[str]) -> bool:
        collection = DataCollections(self.auth).concept_id(concept_id).get()
        if len(collection) > 0 and "s3-links" in collection[0]["meta"]:
            return True
        return False

    def _own_s3_credentials(self, links: List[Dict[str, Any]]) -> Union[str, None]:
        for link in links:
            if "/s3credentials" in link["URL"]:
                return link["URL"]
        return None

    def _running_in_us_west_2(self) -> bool:
        session = self.auth.get_session()
        try:
            # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
            token_ = session.put(
                "http://169.254.169.254/latest/api/token",
                headers={"X-aws-ec2-metadata-token-ttl-seconds": "21600"},
                timeout=1,
            )
            resp = session.get(
                "http://169.254.169.254/latest/meta-data/placement/region",
                timeout=1,
                headers={"X-aws-ec2-metadata-token": token_.text},
            )
        except Exception:
            return False

        if resp.status_code == 200 and b"us-west-2" == resp.content:
            # On AWS, in region us-west-2
            return True
        return False

    def set_requests_session(self, url: str, method: str = "get") -> None:
        """Sets up a `requests` session with bearer tokens that are used by CMR.

        Mainly used to get the authentication cookies from different DAACs and URS.
        This HTTPS session can be used to download granules if we want to use a direct,
        lower level API.

        Parameters:
            url: used to test the credentials and populate the class auth cookies
            method: HTTP method to test, default: "GET"
            bearer_token: if true, will be used for authenticated queries on CMR
        """
        if not hasattr(self, "_http_session"):
            self._http_session = self.auth.get_session()

        resp = self._http_session.request(method, url, allow_redirects=True)

        if resp.status_code in [400, 401, 403]:
            new_session = requests.Session()
            resp_req = new_session.request(
                method, url, allow_redirects=True, cookies=self._requests_cookies
            )
            if resp_req.status_code in [400, 401, 403]:
                resp.raise_for_status()
            else:
                self._requests_cookies.update(new_session.cookies.get_dict())
        elif 200 <= resp.status_code < 300:
            self._requests_cookies = self._http_session.cookies.get_dict()
        else:
            resp.raise_for_status()

    @deprecated("Use get_s3_filesystem instead")
    def get_s3fs_session(
        self,
        daac: Optional[str] = None,
        concept_id: Optional[str] = None,
        provider: Optional[str] = None,
        endpoint: Optional[str] = None,
    ) -> s3fs.S3FileSystem:
        """Returns a s3fs instance for a given cloud provider / DAAC.

        Parameters:
           daac: any of the DAACs, e.g. NSIDC, PODAAC
           provider: a data provider if we know them, e.g. PODAAC -> POCLOUD
           endpoint: pass the URL for the credentials directly

        Returns:
           An `s3fs.S3FileSystem` authenticated for reading in-region in us-west-2 for 1 hour.
        """
        return self.get_s3_filesystem(daac, concept_id, provider, endpoint)

    def get_s3_filesystem(
        self,
        daac: Optional[str] = None,
        concept_id: Optional[str] = None,
        provider: Optional[str] = None,
        endpoint: Optional[str] = None,
    ) -> s3fs.S3FileSystem:
        """Return an `s3fs.S3FileSystem` instance for a given cloud provider / DAAC.

        Parameters:
            daac: any of the DAACs, e.g. NSIDC, PODAAC
            provider: a data provider if we know them, e.g. PODAAC -> POCLOUD
            endpoint: pass the URL for the credentials directly

        Returns:
            a s3fs file instance
        """
        if self.auth is None:
            raise ValueError(
                "A valid Earthdata login instance is required to retrieve S3 credentials"
            )
        if not any([concept_id, daac, provider, endpoint]):
            raise ValueError(
                "At least one of the concept_id, daac, provider or endpoint"
                "parameters must be specified. "
            )

        if concept_id is not None:
            provider = self._derive_concept_provider(concept_id)

        # Get existing S3 credentials if we already have them
        location = (
            daac,
            provider,
            endpoint,
        )  # Identifier for where to get S3 credentials from
        need_new_creds = False
        try:
            dt_init, creds = self._s3_credentials[location]
        except KeyError:
            need_new_creds = True
        else:
            # If cached credentials are expired, invalidate the cache
            delta = datetime.datetime.now() - dt_init
            if round(delta.seconds / 60, 2) > 55:
                need_new_creds = True
                self._s3_credentials.pop(location)

        if need_new_creds:
            # Don't have existing valid S3 credentials, so get new ones
            now = datetime.datetime.now()
            if endpoint is not None:
                creds = self.auth.get_s3_credentials(endpoint=endpoint)
            elif daac is not None:
                creds = self.auth.get_s3_credentials(daac=daac)
            elif provider is not None:
                creds = self.auth.get_s3_credentials(provider=provider)
            # Include new credentials in the cache
            self._s3_credentials[location] = now, creds

        return s3fs.S3FileSystem(
            key=creds["accessKeyId"],
            secret=creds["secretAccessKey"],
            token=creds["sessionToken"],
        )

    @lru_cache
    def get_fsspec_session(self) -> fsspec.AbstractFileSystem:
        """Returns a fsspec HTTPS session with bearer tokens that are used by CMR.

        This HTTPS session can be used to download granules if we want to use a direct,
        lower level API.

        Returns:
            fsspec HTTPFileSystem (aiohttp client session)
        """
        token = self.auth.token["access_token"]
        client_kwargs = {
            "headers": {"Authorization": f"Bearer {token}"},
            # This is important! If we trust the env and send a bearer token,
            # auth will fail!
            "trust_env": False,
        }
        session = fsspec.filesystem("https", client_kwargs=client_kwargs)
        return session

    def get_requests_session(self) -> requests.Session:
        """Returns a requests HTTPS session with bearer tokens that are used by CMR.

        This HTTPS session can be used to download granules if we want to use a direct,
        lower level API.

        Returns:
            requests Session
        """
        if hasattr(self, "_http_session"):
            return self._http_session
        else:
            raise AttributeError("The requests session hasn't been set up yet.")

    def open(
        self,
        granules: Union[List[str], List[DataGranule]],
        provider: Optional[str] = None,
        *,
        show_progress: Optional[bool] = None,
        credentials_endpoint: Optional[str] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
        open_kwargs: Optional[Dict[str, Any]] = None,
    ) -> List[fsspec.spec.AbstractBufferedFile]:
        """Returns a list of file-like objects that can be used to access files
        hosted on S3 or HTTPS by third party libraries like xarray.

        Parameters:
            granules: a list of granule instances **or** list of URLs, e.g. `s3://some-granule`.
                If a list of URLs is passed, we need to specify the data provider.
            provider: e.g. POCLOUD, NSIDC_CPRD, etc.
            show_progress: whether or not to display a progress bar. If not specified, defaults to `True` for interactive sessions
                (i.e., in a notebook or a python REPL session), otherwise `False`.
            credentials_endpoint: S3 credentials endpoint
            pqdm_kwargs: Additional keyword arguments to pass to pqdm, a parallel processing library.
                See pqdm documentation for available options. Default is to use immediate exception behavior.
            open_kwargs: Additional keyword arguments to pass to `fsspec.open`, such as `cache_type` and `block_size`.
                Defaults to using `blockcache` with a block size determined by the file size (4 to 16MB).

        Returns:
            A list of "file pointers" to remote (i.e. `s3://` or `https://`) files.
        """
        if show_progress is None:
            show_progress = _is_interactive()

        pqdm_kwargs = {
            "exception_behaviour": "immediate",
            "n_jobs": 8,
            "disable": not show_progress,
            **(pqdm_kwargs or {}),
        }
        if len(granules):
            return self._open(
                granules,
                provider,
                credentials_endpoint=credentials_endpoint,
                pqdm_kwargs=pqdm_kwargs,
                open_kwargs=open_kwargs,
            )
        return []

    @singledispatchmethod
    def _open(
        self,
        granules: Union[List[str], List[DataGranule]],
        provider: Optional[str] = None,
        *,
        credentials_endpoint: Optional[str] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
        open_kwargs: Optional[Dict[str, Any]] = None,
    ) -> List[Any]:
        raise NotImplementedError("granules should be a list of DataGranule or URLs")

    @_open.register
    def _open_granules(
        self,
        granules: List[DataGranule],
        provider: Optional[str] = None,
        *,
        credentials_endpoint: Optional[str] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
        open_kwargs: Optional[Dict[str, Any]] = None,
    ) -> List[Any]:
        fileset: List = []
        total_size = round(sum([granule.size() for granule in granules]) / 1024, 2)
        logger.info(f"Opening {len(granules)} granules, approx size: {total_size} GB")

        if self.auth is None:
            raise ValueError(
                "A valid Earthdata login instance is required to retrieve credentials"
            )

        if self.in_region:
            if granules[0].cloud_hosted:
                access = "direct"
                provider = granules[0]["meta"]["provider-id"]
                endpoint = credentials_endpoint or self._own_s3_credentials(
                    granules[0]["umm"]["RelatedUrls"]
                )
                if endpoint is not None:
                    logger.info(f"using endpoint: {endpoint}")
                    s3_fs = self.get_s3_filesystem(endpoint=endpoint)
                else:
                    logger.info(f"using provider: {provider}")
                    s3_fs = self.get_s3_filesystem(provider=provider)
            else:
                access = "on_prem"
                s3_fs = None

            url_mapping = _get_url_granule_mapping(granules, access)
            if s3_fs is not None:
                try:
                    fileset = _open_files(
                        url_mapping,
                        fs=s3_fs,
                        pqdm_kwargs=pqdm_kwargs,
                        open_kwargs=open_kwargs,
                    )
                except Exception as e:
                    raise RuntimeError(
                        "An exception occurred while trying to access remote files on S3. "
                        "This may be caused by trying to access the data outside the us-west-2 region."
                        f"Exception: {traceback.format_exc()}"
                    ) from e
            else:
                fileset = self._open_urls_https(
                    url_mapping, pqdm_kwargs=pqdm_kwargs, open_kwargs=open_kwargs
                )
        else:
            url_mapping = _get_url_granule_mapping(granules, access="on_prem")
            fileset = self._open_urls_https(
                url_mapping, pqdm_kwargs=pqdm_kwargs, open_kwargs=open_kwargs
            )

        return fileset

    @_open.register
    def _open_urls(
        self,
        granules: List[str],
        provider: Optional[str] = None,
        *,
        credentials_endpoint: Optional[str] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
        open_kwargs: Optional[Dict[str, Any]] = None,
    ) -> List[Any]:
        fileset: List = []
        s3_fs = None
        if isinstance(granules[0], str) and (
            granules[0].startswith("s3") or granules[0].startswith("http")
        ):
            # TODO: method to derive the DAAC from url?
            provider = provider
        else:
            raise ValueError(
                f"Schema for {granules[0]} is not recognized, must be an HTTP or S3 URL"
            )
        if self.auth is None:
            raise ValueError(
                "A valid Earthdata login instance is required to retrieve S3 credentials"
            )

        url_mapping: Mapping[str, None] = {url: None for url in granules}
        if self.in_region and granules[0].startswith("s3"):
            if provider is not None:
                s3_fs = self.get_s3_filesystem(provider=provider)
            elif credentials_endpoint is not None:
                s3_fs = self.get_s3_filesystem(endpoint=credentials_endpoint)
            if s3_fs:
                try:
                    fileset = _open_files(
                        url_mapping,
                        fs=s3_fs,
                        pqdm_kwargs=pqdm_kwargs,
                        open_kwargs=open_kwargs,
                    )
                except Exception as e:
                    raise RuntimeError(
                        "An exception occurred while trying to access remote files on S3. "
                        "This may be caused by trying to access the data outside the us-west-2 region. "
                        f"Exception: {traceback.format_exc()}"
                    ) from e

                return fileset
            else:
                logger.error(
                    f"An error occurred while trying to retrieve the cloud credentials for provider: {provider}. endpoint: {credentials_endpoint}"
                )
                return fileset
        else:
            if granules[0].startswith("s3"):
                raise ValueError(
                    "We cannot open S3 links when we are not in-region, try using HTTPS links"
                )
            fileset = self._open_urls_https(url_mapping, pqdm_kwargs=pqdm_kwargs)
            return fileset

    def get(
        self,
        granules: Union[List[DataGranule], List[str]],
        local_path: Optional[Union[Path, str]] = None,
        provider: Optional[str] = None,
        threads: int = 8,
        *,
        credentials_endpoint: Optional[str] = None,
        show_progress: Optional[bool] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> List[Path]:
        """Retrieves data granules from a remote storage system.

           * If we run this in the cloud,
             we are moving data from S3 to a cloud compute instance (EC2, AWS Lambda).
           * If we run it outside the us-west-2 region and the data granules are part of a cloud-based
             collection, the method will not get any files.
           * If we request data granules from an on-prem collection,
             the data will be effectively downloaded to a local directory.

        Parameters:
            granules: A list of granules(DataGranule) instances or a list of granule links (HTTP).
            local_path: Local directory to store the remote data granules.  If not
                supplied, defaults to a subdirectory of the current working directory
                of the form `data/YYYY-MM-DD-UUID`, where `YYYY-MM-DD` is the year,
                month, and day of the current date, and `UUID` is the last 6 digits
                of a UUID4 value.
            provider: a valid cloud provider, each DAAC has a provider code for their cloud distributions
            credentials_endpoint: If provided, this will be used to get S3 credentials
            threads: Parallel number of threads to use to download the files;
                adjust as necessary, default = 8.
            show_progress: whether or not to display a progress bar. If not specified, defaults to `True` for interactive sessions
                (i.e., in a notebook or a python REPL session), otherwise `False`.
            pqdm_kwargs: Additional keyword arguments to pass to pqdm, a parallel processing library.
                See pqdm documentation for available options. Default is to use immediate exception behavior
                and the number of jobs specified by the `threads` parameter.

        Returns:
            List of downloaded files
        """
        if not granules:
            raise ValueError("List of URLs or DataGranule instances expected")

        if local_path is None:
            today = datetime.datetime.now().strftime("%Y-%m-%d")
            uuid = uuid4().hex[:6]
            local_path = Path.cwd() / "data" / f"{today}-{uuid}"

        if show_progress is None:
            show_progress = _is_interactive()

        pqdm_kwargs = {
            "exception_behaviour": "immediate",  # should be overridden by pqdm_kwargs if passed
            "n_jobs": threads,
            "disable": not show_progress,
            **(pqdm_kwargs or {}),
        }

        return self._get(
            granules,
            Path(local_path),
            provider,
            credentials_endpoint=credentials_endpoint,
            pqdm_kwargs=pqdm_kwargs,
        )

    @singledispatchmethod
    def _get(
        self,
        granules: Union[List[DataGranule], List[str]],
        local_path: Path,
        provider: Optional[str] = None,
        *,
        credentials_endpoint: Optional[str] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> List[Path]:
        """Retrieves data granules from a remote storage system.

           * If we run this in the cloud,
             we are moving data from S3 to a cloud compute instance (EC2, AWS Lambda).
           * If we run it outside the us-west-2 region and the data granules are part of a cloud-based
             collection, the method will not get any files.
           * If we request data granules from an on-prem collection,
             the data will be effectively downloaded to a local directory.

        Parameters:
            granules: A list of granules (DataGranule) instances or a list of granule links (HTTP).
            local_path: Local directory to store the remote data granules
            provider: a valid cloud provider, each DAAC has a provider code for their cloud distributions
            pqdm_kwargs: Additional keyword arguments to pass to pqdm, a parallel processing library.
                See pqdm documentation for available options. Default is to use immediate exception behavior
                and the number of jobs specified by the `threads` parameter.

        Returns:
            None
        """
        raise NotImplementedError(f"Cannot _get {granules}")

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=1, max=10),
        retry=retry_if_exception_type(Exception),
    )
    def download_cloud_file(
        self, s3_fs: fsspec.AbstractFileSystem, file: str, local_path: Path
    ) -> Path:
        file_name = local_path / Path(file).name
        if file_name.exists():
            return file_name  # Skip if already exists

        s3_fs.get([file], str(local_path), recursive=False)
        logger.info(f"Downloading: {file_name}")
        return file_name

    @_get.register
    def _get_urls(
        self,
        granules: List[str],
        local_path: Path,
        provider: Optional[str] = None,
        *,
        credentials_endpoint: Optional[str] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> List[Path]:
        data_links = granules
        s3_fs = s3fs.S3FileSystem()
        if (
            provider is None
            and credentials_endpoint is None
            and self.in_region
            and "cumulus" in data_links[0]
        ):
            raise ValueError(
                "earthaccess can't yet guess the provider for cloud collections, "
                "we need to use one from `earthaccess.list_cloud_providers()` or if known the S3 credential endpoint"
            )
        if self.in_region and data_links[0].startswith("s3"):
            if credentials_endpoint is not None:
                logger.info(
                    f"Accessing cloud dataset using credentials_endpoint: {credentials_endpoint}"
                )
                s3_fs = self.get_s3_filesystem(endpoint=credentials_endpoint)
            elif provider is not None:
                logger.info(f"Accessing cloud dataset using provider: {provider}")
                s3_fs = self.get_s3_filesystem(provider=provider)

            def _download(file: str) -> Union[Path, None]:
                return self.download_cloud_file(s3_fs, file, local_path)

            results = pqdm(data_links, _download, **pqdm_kwargs)
            return [r for r in results if r is not None]

        else:
            # if we are not in AWS
            return self._download_onprem_granules(
                data_links, local_path, pqdm_kwargs=pqdm_kwargs
            )

    @_get.register
    def _get_granules(
        self,
        granules: List[DataGranule],
        local_path: Path,
        provider: Optional[str] = None,
        *,
        credentials_endpoint: Optional[str] = None,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> List[Path]:
        data_links: List = []
        provider = granules[0]["meta"]["provider-id"]
        endpoint = self._own_s3_credentials(granules[0]["umm"]["RelatedUrls"])
        cloud_hosted = granules[0].cloud_hosted
        access = "direct" if (cloud_hosted and self.in_region) else "external"
        data_links = list(
            # we are not in-region
            chain.from_iterable(
                granule.data_links(access=access, in_region=self.in_region)
                for granule in granules
            )
        )
        total_size = round(sum(granule.size() for granule in granules) / 1024, 2)
        logger.info(
            f" Getting {len(granules)} granules, approx download size: {total_size} GB"
        )
        if access == "direct":
            if endpoint is not None:
                logger.info(
                    f"Accessing cloud dataset using dataset endpoint credentials: {endpoint}"
                )
                s3_fs = self.get_s3_filesystem(endpoint=endpoint)
            else:
                logger.info(f"Accessing cloud dataset using provider: {provider}")
                s3_fs = self.get_s3_filesystem(provider=provider)

            local_path.mkdir(parents=True, exist_ok=True)

            def _download(file: str) -> Union[Path, None]:
                return self.download_cloud_file(s3_fs, file, local_path)

            results = pqdm(data_links, _download, **pqdm_kwargs)
            return [r for r in results if r is not None]

        else:
            # if the data are cloud-based, but we are not in AWS,
            # it will be downloaded as if it was on prem
            return self._download_onprem_granules(
                data_links, local_path, pqdm_kwargs=pqdm_kwargs
            )

    def _clone_session_in_local_thread(
        self, original_session: requests.Session
    ) -> None:
        """Clone the original session and store it in the local thread context.

        This method creates a new session that replicates the headers, cookies, and authentication settings
        from the provided original session. The new session is stored in a thread-local storage.

        Parameters:
            original_session (SessionWithHeaderRedirection): The session to be cloned.

        Returns:
            None
        """
        if not hasattr(self.thread_locals, "local_thread_session"):
            local_thread_session = self.auth.get_session()  # type: ignore
            local_thread_session.headers.update(original_session.headers)
            local_thread_session.cookies.update(original_session.cookies)
            local_thread_session.auth = original_session.auth
            self.thread_locals.local_thread_session = local_thread_session

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=1, max=10),
        retry=retry_if_exception_type(Exception),
    )
    def _download_file(self, url: str, directory: Path) -> Path:
        """Download a single file using a bearer token.

        Parameters:
            url: the granule url
            directory: local directory

        Returns:
            A local filepath or an exception.
        """
        # If the get data link is an Opendap location
        if "opendap" in url and url.endswith(".html"):
            url = url.replace(".html", "")
        local_filename = url.split("/")[-1]
        path = directory / Path(local_filename)
        if not path.exists():
            original_session = self.get_requests_session()
            # This reuses the auth cookie, we make sure we only authenticate N threads instead
            # of one per file, see #913
            self._clone_session_in_local_thread(original_session)
            session = self.thread_locals.local_thread_session
            with session.get(url, stream=True, allow_redirects=True) as r:
                if r.status_code in [401, 403]:
                    text = (r.text or "").lower()
                    if "eula" in text:
                        raise EulaNotAccepted(f"Eula Acceptance Failure for {url}")
                if r.status_code >= 400:
                    raise DownloadFailure(
                        f"Download failed for {url}. Status code: {r.status_code}"
                    )

                with open(path, "wb") as f:
                    # Cap memory usage for large files at 1MB per write to disk per thread
                    # https://docs.python-requests.org/en/latest/user/quickstart/#raw-response-content
                    for chunk in r.iter_content(chunk_size=1024 * 1024):
                        f.write(chunk)
        else:
            logger.info(f"File {local_filename} already downloaded")
        return path

    def _download_onprem_granules(
        self,
        urls: List[str],
        directory: Path,
        *,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> List[Any]:
        """Downloads a list of URLS into the data directory.

        Parameters:
            urls: list of granule URLs from an on-prem collection
            directory: local directory to store the downloaded files
            pqdm_kwargs: Additional keyword arguments to pass to pqdm, a parallel processing library.
                See pqdm documentation for available options. Default is to use immediate exception behavior
                and the number of jobs specified by the `threads` parameter.

        Returns:
            A list of local filepaths to which the files were downloaded.
        """
        if urls is None:
            raise ValueError("The granules didn't provide a valid GET DATA link")
        if self.auth is None:
            raise ValueError(
                "We need to be logged into NASA EDL in order to download data granules"
            )
        directory.mkdir(parents=True, exist_ok=True)

        arguments = [(url, directory) for url in urls]

        pqdm_kwargs = {
            "exception_behaviour": "immediate",
            "disable": True,
            **(pqdm_kwargs or {}),
            # We don't want a user to be able to override the following kwargs,
            # which is why they appear *after* spreading pqdm_kwargs above.
            "argument_type": "args",
        }

        return pqdm(arguments, self._download_file, **pqdm_kwargs)

    def _open_urls_https(
        self,
        url_mapping: Mapping[str, Union[DataGranule, None]],
        *,
        pqdm_kwargs: Optional[Mapping[str, Any]] = None,
        open_kwargs: Optional[Dict[str, Any]] = None,
    ) -> List[fsspec.AbstractFileSystem]:
        https_fs = self.get_fsspec_session()

        try:
            return _open_files(
                url_mapping, https_fs, pqdm_kwargs=pqdm_kwargs, open_kwargs=open_kwargs
            )
        except Exception:
            logger.exception(
                "An exception occurred while trying to access remote files via HTTPS"
            )
            raise



import sys
from datetime import date, datetime
from typing import Any, Optional, SupportsFloat, Union

if sys.version_info < (3, 9):
    from typing import List, MutableMapping, Sequence, Tuple
else:
    from builtins import list as List, tuple as Tuple
    from collections.abc import MutableMapping, Sequence

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

CMR_OPS: str
CMR_UAT: str
CMR_SIT: str

FloatLike: TypeAlias = Union[str, SupportsFloat]
PointLike: TypeAlias = Tuple[FloatLike, FloatLike]

class Query:
    params: MutableMapping[str, Any]
    options: MutableMapping[str, Any]
    concept_id_chars: Sequence[str]
    headers: MutableMapping[str, str]

    def __init__(self, route: str, mode: str = ...) -> None: ...
    def _build_url(self) -> str: ...
    def get(self, limit: int = ...) -> List[Any]: ...
    def hits(self) -> int: ...
    def get_all(self) -> List[Any]: ...
    def parameters(self, **kwargs: Any) -> Self: ...
    def format(self, output_format: str = "json") -> Self: ...
    def concept_id(self, ids: Sequence[str]) -> Self: ...
    def provider(self, provider: str) -> Self: ...
    def mode(self, mode: str = ...) -> None: ...
    def token(self, token: str) -> Self: ...
    def bearer_token(self, bearer_token: str) -> Self: ...

class GranuleCollectionBaseQuery(Query):
    def online_only(self, online_only: bool = True) -> Self: ...
    def temporal(
        self,
        date_from: Optional[Union[str, date, datetime]],
        date_to: Optional[Union[str, date, datetime]],
        exclude_boundary: bool = False,
    ) -> Self: ...
    def short_name(self, short_name: str) -> Self: ...
    def version(self, version: str) -> Self: ...
    def point(self, lon: FloatLike, lat: FloatLike) -> Self: ...
    def circle(self, lon: FloatLike, lat: FloatLike, dist: FloatLike) -> Self: ...
    def polygon(self, coordinates: Sequence[PointLike]) -> Self: ...
    def bounding_box(
        self,
        lower_left_lon: FloatLike,
        lower_left_lat: FloatLike,
        upper_right_lon: FloatLike,
        upper_right_lat: FloatLike,
    ) -> Self: ...
    def line(self, coordinates: Sequence[PointLike]) -> Self: ...
    def downloadable(self, downloadable: bool = True) -> Self: ...
    def entry_title(self, entry_title: str) -> Self: ...

class GranuleQuery(GranuleCollectionBaseQuery):
    def __init__(self, mode: str = ...) -> None: ...
    def orbit_number(
        self,
        orbit1: FloatLike,
        orbit2: Optional[FloatLike] = ...,
    ) -> Self: ...
    def day_night_flag(self, day_night_flag: str) -> Self: ...
    def cloud_cover(
        self,
        min_cover: Optional[FloatLike] = ...,
        max_cover: Optional[FloatLike] = ...,
    ) -> Self: ...
    def instrument(self, instrument: str) -> Self: ...
    def platform(self, platform: str) -> Self: ...
    def sort_key(self, sort_key: str) -> Self: ...
    def granule_ur(self, granule_ur: str) -> Self: ...

class CollectionQuery(GranuleCollectionBaseQuery):
    def __init__(self, mode: str = ...) -> None: ...
    def archive_center(self, center: str) -> Self: ...
    def keyword(self, text: str) -> Self: ...
    def native_id(self, native_ids: Sequence[str]) -> Self: ...
    def tool_concept_id(self, ids: Sequence[str]) -> Self: ...
    def service_concept_id(self, ids: Sequence[str]) -> Self: ...

class ToolServiceVariableBaseQuery(Query):
    def native_id(self, native_ids: Sequence[str]) -> Self: ...
    def name(self, name: str) -> Self: ...

class ToolQuery(ToolServiceVariableBaseQuery):
    def __init__(self, mode: str = ...) -> None: ...

class ServiceQuery(ToolServiceVariableBaseQuery):
    def __init__(self, mode: str = ...) -> None: ...

class VariableQuery(ToolServiceVariableBaseQuery):
    def __init__(self, mode: str = ...) -> None: ...



# package imports
import os
import threading
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

import fsspec
import pytest
import responses
import s3fs
from earthaccess import Auth, Store
from earthaccess.auth import SessionWithHeaderRedirection
from earthaccess.exceptions import DownloadFailure, EulaNotAccepted
from earthaccess.store import EarthAccessFile, _open_files
from pqdm.threads import pqdm


class TestEula(unittest.TestCase):
    @patch.dict(
        os.environ,
        {
            "EARTHDATA_USERNAME": "user_no_eula",
            "EARTHDATA_PASSWORD": "password",
        },
        clear=True,
    )
    @responses.activate
    def setUp(self):
        json_response = {"access_token": "EDL-token-1", "expiration_date": "12/15/2021"}
        responses.add(
            responses.GET,
            "https://urs.earthdata.nasa.gov/profile",
            json=json_response,
            status=200,
        )
        responses.add(
            responses.POST,
            "https://urs.earthdata.nasa.gov/api/users/find_or_create_token",
            json=json_response,
            status=200,
        )

        self.auth = Auth()
        self.auth.login(strategy="environment")
        assert self.auth.authenticated

    @responses.activate
    def test_eula_detects_401_errors(self):
        response = " blah blah Eula bing!"
        mocked_url = "https://example.com/protected_file.nc"
        responses.add(
            responses.GET,
            "https://urs.earthdata.nasa.gov/profile",
            json={},
            status=200,
        )
        responses.add(
            responses.GET,
            url=mocked_url,
            body=response,
            status=401,
        )
        store = Store(self.auth)
        with self.assertRaisesRegex(
            EulaNotAccepted, f"Eula Acceptance Failure for {mocked_url}"
        ):
            store.get([mocked_url], "/tmp")

    @responses.activate
    def test_detects_non_eula_errors(self):
        response = " blah blah error!"
        mocked_url = "https://example.com/protected_file.nc"
        responses.add(
            responses.GET,
            "https://urs.earthdata.nasa.gov/profile",
            json={},
            status=200,
        )
        responses.add(
            responses.GET,
            url=mocked_url,
            body=response,
            status=401,
        )
        store = Store(self.auth)
        with self.assertRaisesRegex(
            DownloadFailure, f"Download failed for {mocked_url}. Status code: 401"
        ):
            store.get([mocked_url], "/tmp")

    def tearDown(self):
        self.auth = None

    @responses.activate
    def test_store_can_create_https_fsspec_session(self):
        responses.add(
            responses.GET,
            "https://urs.earthdata.nasa.gov/profile",
            json={},
            status=200,
        )
        store = Store(self.auth)
        assert isinstance(store.auth, Auth)
        https_fs = store.get_fsspec_session()
        assert type(https_fs) is type(fsspec.filesystem("https"))
        return None


class TestStoreSessions(unittest.TestCase):
    @responses.activate
    @patch.dict(
        os.environ,
        {
            "EARTHDATA_USERNAME": "user",
            "EARTHDATA_PASSWORD": "password",
        },
        clear=True,
    )
    def setUp(self):
        json_response = {"access_token": "EDL-token-1", "expiration_date": "12/15/2021"}
        responses.add(
            responses.POST,
            "https://urs.earthdata.nasa.gov/api/users/find_or_create_token",
            json=json_response,
            status=200,
        )
        self.auth = Auth()
        self.auth.login(strategy="environment")
        self.assertEqual(self.auth.authenticated, True)
        self.assertEqual(self.auth.token, json_response)

    def tearDown(self):
        self.auth = None

    @responses.activate
    def test_store_can_create_https_fsspec_session(self):
        responses.add(
            responses.GET,
            "https://urs.earthdata.nasa.gov/profile",
            json={},
            status=200,
        )
        store = Store(self.auth)
        self.assertTrue(isinstance(store.auth, Auth))
        https_fs = store.get_fsspec_session()
        self.assertEqual(type(https_fs), type(fsspec.filesystem("https")))
        return None

    @responses.activate
    def test_store_can_create_s3_fsspec_session(self):
        from earthaccess.daac import DAACS

        custom_endpoints = [
            "https://archive.swot.podaac.earthdata.nasa.gov/s3credentials",
            "https://api.giovanni.earthdata.nasa.gov/s3credentials",
            "https://data.laadsdaac.earthdatacloud.nasa.gov/s3credentials",
        ]
        mock_creds = {
            "accessKeyId": "sure",
            "secretAccessKey": "correct",
            "sessionToken": "whynot",
        }
        expected_storage_options = {
            "key": mock_creds["accessKeyId"],
            "secret": mock_creds["secretAccessKey"],
            "token": mock_creds["sessionToken"],
        }

        for endpoint in custom_endpoints:
            responses.add(
                responses.GET,
                endpoint,
                json=mock_creds,
                status=200,
            )

        for daac in DAACS:
            if "s3-credentials" in daac:
                responses.add(
                    responses.GET,
                    daac["s3-credentials"],
                    json=mock_creds,
                    status=200,
                )
        responses.add(
            responses.GET,
            "https://urs.earthdata.nasa.gov/profile",
            json=mock_creds,
            status=200,
        )

        store = Store(self.auth)
        self.assertTrue(isinstance(store.auth, Auth))
        for daac in [
            "NSIDC",
            "PODAAC",
            "LPDAAC",
            "ORNLDAAC",
            "GES_DISC",
            "ASF",
            "OBDAAC",
            "ASDC",
        ]:
            s3_fs = store.get_s3_filesystem(daac=daac)
            assert isinstance(s3_fs, s3fs.S3FileSystem)
            assert s3_fs.storage_options == expected_storage_options

        for endpoint in custom_endpoints:
            s3_fs = store.get_s3_filesystem(endpoint=endpoint)
            assert isinstance(s3_fs, s3fs.S3FileSystem)
            assert s3_fs.storage_options == expected_storage_options

        for provider in [
            "NSIDC_CPRD",
            "POCLOUD",
            "LPCLOUD",
            "ORNL_CLOUD",
            "GES_DISC",
            "ASF",
            "OB_CLOUD",
            "LARC_CLOUD",
        ]:
            s3_fs = store.get_s3_filesystem(provider=provider)
            assert isinstance(s3_fs, s3fs.S3FileSystem)
            assert s3_fs.storage_options == expected_storage_options

        # Ensure informative error is raised
        with pytest.raises(ValueError, match="parameters must be specified"):
            store.get_s3_filesystem()

        return None

    @responses.activate
    def test_session_reuses_token_download(self):
        mock_creds = {
            "accessKeyId": "sure",
            "secretAccessKey": "correct",
            "sessionToken": "whynot",
        }
        test_cases = [
            (2, 500),  # 2 threads, 500 files
            (4, 400),  # 4 threads, 400 files
            (8, 5000),  # 8 threads, 5k files
        ]
        for n_threads, n_files in test_cases:
            with self.subTest(n_threads=n_threads, n_files=n_files):
                urls = [f"https://example.com/file{i}" for i in range(1, n_files + 1)]
                for i, url in enumerate(urls):
                    responses.add(
                        responses.GET, url, body=f"Content of file {i + 1}", status=200
                    )

                edl_hostname = "urs.earthdata.nasa.gov"
                mock_auth = MagicMock()
                mock_auth.authenticated = True
                mock_auth.system.edl_hostname = edl_hostname
                responses.add(
                    responses.GET,
                    "https://urs.earthdata.nasa.gov/profile",
                    json=mock_creds,
                    status=200,
                )

                original_session = SessionWithHeaderRedirection(edl_hostname)
                original_session.cookies.set("sessionid", "mocked-session-cookie")
                mock_auth.get_session.return_value = original_session

                store = Store(auth=mock_auth)
                store.thread_locals = threading.local()  # Use real thread-local storage

                # Track cloned sessions
                cloned_sessions = set()

                def mock_clone_session_in_local_thread(original_session):
                    """Mock session cloning to track cloned sessions."""
                    if not hasattr(store.thread_locals, "local_thread_session"):
                        session = SessionWithHeaderRedirection(edl_hostname)
                        session.cookies.update(original_session.cookies)
                        cloned_sessions.add(id(session))
                        store.thread_locals.local_thread_session = session

                with patch.object(
                    store,
                    "_clone_session_in_local_thread",
                    side_effect=mock_clone_session_in_local_thread,
                ):
                    mock_directory = Path("/mock/directory")
                    downloaded_files = []

                    def mock_download_file(url):
                        """Mock file download to track downloaded files."""
                        # Ensure session cloning happens before downloading
                        store._clone_session_in_local_thread(original_session)
                        downloaded_files.append(url)
                        return mock_directory / f"{url.split('/')[-1]}"

                    with patch.object(
                        store, "_download_file", side_effect=mock_download_file
                    ):
                        # Test multi-threaded download
                        pqdm(urls, store._download_file, n_jobs=n_threads)  # type: ignore

                # We make sure we reuse the token up to N threads
                self.assertTrue(len(cloned_sessions) <= n_threads)

                self.assertEqual(len(downloaded_files), n_files)  # 10 files downloaded
                self.assertCountEqual(downloaded_files, urls)  # All files accounted for


def test_earthaccess_file_getattr():
    fs = fsspec.filesystem("memory")
    with fs.open("foo", "wb") as f:
        earthaccess_file = EarthAccessFile(f, granule="foo")
        assert f.tell == earthaccess_file.tell
    fs.store.clear()


@pytest.mark.parametrize(
    "file_size, open_kwargs, expected_cache_type, expected_block_size",
    [
        # Case 1: Small file, defaults used
        (50 * 1024 * 1024, {}, "background", 4 * 1024 * 1024),
        # Case 2: Medium file, block_size scales up
        (300 * 1024 * 1024, {}, "background", 8 * 1024 * 1024),
        # Case 3: Large file, hits max block size
        (1600 * 1024 * 1024, {}, "background", 16 * 1024 * 1024),
        # Case 4: Override cache_type and block_size
        (
            600 * 1024 * 1024,
            {"cache_type": "readahead", "block_size": 1024},
            "readahead",
            1024,
        ),
        # Case 5: Override only one (block_size)
        (
            600 * 1024 * 1024,
            {"block_size": 8 * 1024 * 1024},
            "background",
            8 * 1024 * 1024,
        ),
    ],
)
def test_open_files_parametrized(
    file_size, open_kwargs, expected_cache_type, expected_block_size
):
    fs = MagicMock()
    fs.info.return_value = {"size": file_size}
    fs.open = MagicMock()

    url = "https://example.com/file.nc"
    granule = MagicMock()
    url_mapping = {url: granule}

    result = _open_files(url_mapping, fs, open_kwargs=open_kwargs)

    fs.open.assert_called_once()
    args, kwargs = fs.open.call_args

    assert args[0] == url
    assert result[0] is not None
    assert kwargs["cache_type"] == expected_cache_type
    assert kwargs["block_size"] == expected_block_size or (
        isinstance(expected_block_size, float)
        and abs(kwargs["block_size"] - expected_block_size) < 1e5
    )



from __future__ import annotations

from typing import Optional, Union

import fsspec
import fsspec.utils
import s3fs

# import ipdb
import earthaccess


def _get_chunk_metadata(
    granule: earthaccess.DataGranule,
    fs: fsspec.AbstractFileSystem,
) -> list[dict]:
    from kerchunk.hdf import SingleHdf5ToZarr

    if not isinstance(granule, earthaccess.DataGranule) and isinstance(granule, dict):
        # WHY: dask serialization is doing something weird, it serializes the granule as a simple dict
        # we need to add cast it back to a datagranule to get the nice methods for parsing the data links
        # TODO: ask James what is going on
        granule = earthaccess.DataGranule(granule)

    metadata = []
    access = "direct" if isinstance(fs, s3fs.S3FileSystem) else "indirect"
    # ipdb.set_trace()

    for url in granule.data_links(access=access):
        with fs.open(url) as inf:
            h5chunks = SingleHdf5ToZarr(inf, url)  # type: ignore
            m = h5chunks.translate()
            metadata.append(m)

    return metadata


def consolidate_metadata(
    granules: list[earthaccess.DataGranule],
    kerchunk_options: Optional[dict] = None,
    access: str = "direct",
    outfile: Optional[str] = None,
    storage_options: Optional[dict] = None,
) -> Union[str, dict]:
    try:
        import dask
        from kerchunk.combine import MultiZarrToZarr
    except ImportError as e:
        raise ImportError(
            "`earthaccess.consolidate_metadata` requires `dask` and `kerchunk` to be be installed"
        ) from e

    if access == "direct":
        fs = earthaccess.get_s3_filesystem(provider=granules[0]["meta"]["provider-id"])
    else:
        fs = earthaccess.get_fsspec_https_session()

    # Get metadata for each granule
    get_chunk_metadata = dask.delayed(_get_chunk_metadata)  # type: ignore

    # ipdb.set_trace()
    chunks = dask.compute(*[get_chunk_metadata(g, fs) for g in granules])  # type: ignore
    chunks = sum(chunks, start=[])

    # Get combined metadata object
    mzz = MultiZarrToZarr(chunks, **(kerchunk_options or {}))

    if outfile is None:
        return mzz.translate()

    output = fsspec.utils.stringify_path(outfile)
    mzz.translate(outfile, storage_options=storage_options or {})
    return output



import contextlib
import json
import logging
import os.path

import earthaccess
import responses
from earthaccess.results import DataCollection
from earthaccess.search import DataCollections
from vcr.unittest import VCRTestCase  # type: ignore[import-untyped]

logging.basicConfig()
logging.getLogger("vcr").setLevel(logging.ERROR)

REDACTED_STRING = "REDACTED"


def unique_results(results):
    """When we invoke a search request multiple times we want to ensure that we don't
    get the same results back. This is a one shot test as the results are preserved
    by VCR but still useful.
    """
    unique_concept_ids = {result["meta"]["concept-id"] for result in results}
    return len(unique_concept_ids) == len(results)


def redact_login_request(request):
    if "/api/users/" in request.path and "/api/users/tokens" not in request.path:
        _, user_name = os.path.split(request.path)
        request.uri = request.uri.replace(user_name, REDACTED_STRING)

    return request


def redact_key_values(keys_to_redact):
    def redact(payload):
        for key in keys_to_redact:
            if key in payload:
                payload[key] = REDACTED_STRING
        return payload

    def before_record_response(response):
        body = response["body"]["string"].decode("utf8")

        with contextlib.suppress(json.JSONDecodeError):
            payload = json.loads(body)
            redacted_payload = (
                list(map(redact, payload))
                if isinstance(payload, list)
                else redact(payload)
            )
            response["body"]["string"] = json.dumps(redacted_payload).encode()

        return response

    return before_record_response


class TestResults(VCRTestCase):
    def _get_vcr(self, **kwargs):
        myvcr = super()._get_vcr(**kwargs)
        myvcr.cassette_library_dir = "tests/unit/fixtures/vcr_cassettes"
        myvcr.decode_compressed_response = True
        # Header matching is not set by default, we need that to test the
        # search-after functionality is performing correctly.
        myvcr.match_on = [
            "method",
            "scheme",
            "host",
            "port",
            "path",
            "query",
            "headers",
        ]
        myvcr.filter_headers = [
            "Accept-Encoding",
            "Authorization",
            "Cookie",
            "Set-Cookie",
            "User-Agent",
        ]
        myvcr.filter_query_parameters = [
            ("client_id", REDACTED_STRING),
        ]

        myvcr.before_record_response = redact_key_values(
            [
                "access_token",
                "uid",
                "first_name",
                "last_name",
                "email_address",
                "nams_auid",
            ]
        )

        myvcr.before_record_request = redact_login_request

        return myvcr

    def test_no_results(self):
        """If we search for a collection that doesn't exist, we should get no results."""
        granules = earthaccess.search_data(
            # STAC collection name; correct short name is OPERA_L3_DSWX-HLS_V1
            # Example discussed in: https://github.com/nsidc/earthaccess/pull/839
            short_name="OPERA_L3_DSWX-HLS_V1_1.0",
            bounding_box=(-95.19, 30.59, -94.99, 30.79),
            temporal=("2024-04-30", "2024-05-31"),
        )
        assert len(granules) == 0

    def test_data_links(self):
        granules = earthaccess.search_data(
            short_name="SEA_SURFACE_HEIGHT_ALT_GRIDS_L4_2SATS_5DAY_6THDEG_V_JPL2205",
            temporal=("2020", "2022"),
            count=1,
        )

        g = granules[0]
        # `access` specified
        assert g.data_links(access="direct")[0].startswith("s3://")
        assert g.data_links(access="external")[0].startswith("https://")
        # `in_region` specified
        assert g.data_links(in_region=True)[0].startswith("s3://")
        assert g.data_links(in_region=False)[0].startswith("https://")
        # When `access` and `in_region` are both specified, `access` takes priority
        assert g.data_links(access="direct", in_region=True)[0].startswith("s3://")
        assert g.data_links(access="direct", in_region=False)[0].startswith("s3://")
        assert g.data_links(access="external", in_region=True)[0].startswith("https://")
        assert g.data_links(access="external", in_region=False)[0].startswith(
            "https://"
        )

    def test_get_more_than_2000(self):
        """If we execute a get with a limit of more than 2000
        then we expect multiple invocations of a cmr granule search and
        to not fetch back more results than we ask for.
        """
        granules = earthaccess.search_data(short_name="MOD02QKM", count=3000)

        # Assert that we performed one 'hits' search and two 'results' search queries
        self.assertEqual(len(self.cassette), 3)
        self.assertEqual(len(granules), 4000)
        self.assertTrue(unique_results(granules))

    def test_get(self):
        """If we execute a get with no arguments then we expect
        to get the maximum no. of granules from a single CMR call (2000)
        in a single request.
        """
        granules = earthaccess.search_data(short_name="MOD02QKM", count=2000)

        # Assert that we performed one 'hits' search and one 'results' search queries
        self.assertEqual(len(self.cassette), 2)
        self.assertEqual(len(granules), 2000)
        self.assertTrue(unique_results(granules))

    def test_get_all_less_than_2k(self):
        """If we execute a get_all then we expect multiple
        invocations of a cmr granule search and
        to not fetch back more results than we ask for.
        """
        granules = earthaccess.search_data(
            short_name="TELLUS_GRAC_L3_JPL_RL06_LND_v04", count=2000
        )

        # Assert that we performed a hits query and one search results query
        self.assertEqual(len(self.cassette), 2)
        self.assertEqual(len(granules), 163)
        self.assertTrue(unique_results(granules))

    def test_get_all_more_than_2k(self):
        """If we execute a get_all then we expect multiple
        invocations of a cmr granule search and
        to not fetch back more results than we ask for.
        """
        granules = earthaccess.search_data(
            short_name="CYGNSS_NOAA_L2_SWSP_25KM_V1.2", count=3000
        )

        # Assert that we performed a hits query and two search results queries
        self.assertEqual(len(self.cassette), 3)
        self.assertEqual(
            len(granules), int(self.cassette.responses[0]["headers"]["CMR-Hits"][0])
        )
        self.assertEqual(
            len(granules),
            min(3000, int(self.cassette.responses[0]["headers"]["CMR-Hits"][0])),
        )
        self.assertTrue(unique_results(granules))

    def test_collections_less_than_2k(self):
        """If we execute a get_all then we expect multiple
        invocations of a cmr granule search and
        to not fetch back more results than we ask for.
        """
        query = DataCollections().daac("PODAAC").cloud_hosted(True)
        collections = query.get(20)

        # Assert that we performed a single search results query
        self.assertEqual(len(self.cassette), 1)
        self.assertEqual(len(collections), 20)
        self.assertTrue(unique_results(collections))
        self.assert_is_using_search_after(self.cassette)

    def test_collections_more_than_2k(self):
        """If we execute a get_all then we expect multiple
        invocations of a cmr granule search and
        to not fetch back more results than we ask for.
        """
        query = DataCollections()
        collections = query.get(3000)

        # Assert that we performed two search results queries
        self.assertEqual(len(self.cassette), 2)
        self.assertEqual(len(collections), 4000)
        self.assertTrue(unique_results(collections))
        self.assert_is_using_search_after(self.cassette)

    def assert_is_using_search_after(self, cass):
        first_request = True

        for request in cass.requests:
            # Verify the page number was not used
            self.assertTrue("page_num" not in request.uri)
            # Verify that Search After was used in all requests except first
            self.assertEqual(first_request, "CMR-Search-After" not in request.headers)
            first_request = False


def test_get_doi_returns_doi_when_present():
    collection = DataCollection(
        {"umm": {"DOI": {"DOI": "doi:10.16904/envidat.lwf.34"}}, "meta": {}}
    )

    assert collection.doi() == "doi:10.16904/envidat.lwf.34"


def test_get_doi_returns_empty_string_when_doi_missing():
    collection = DataCollection({"umm": {"DOI": {}}, "meta": {}})

    assert collection.doi() is None


def test_get_doi_returns_empty_string_when_doi_key_missing():
    collection = DataCollection({"umm": {}, "meta": {}})

    assert collection.doi() is None


@responses.activate
def test_get_citation_apa_format():
    collection = DataCollection(
        {"umm": {"DOI": {"DOI": "doi:10.16904/envidat.lwf.34"}}, "meta": {}}
    )

    responses.add(
        responses.GET,
        "https://citation.doi.org/format?doi=doi:10.16904/envidat.lwf.34&style=apa&lang=en-US",
        body="Meusburger, K., Graf Pannatier, E., & Schaub, M. (2019). 10-HS Pfynwald (Version 2019) [Dataset]. EnviDat. https://doi.org/10.16904/ENVIDAT.LWF.34",
        status=200,
    )

    citation = collection.citation(format="apa", language="en-US")

    assert (
        citation
        == "Meusburger, K., Graf Pannatier, E., & Schaub, M. (2019). 10-HS Pfynwald (Version 2019) [Dataset]. EnviDat. https://doi.org/10.16904/ENVIDAT.LWF.34"
    )


@responses.activate
def test_get_citation_different_language():
    collection = DataCollection(
        {"umm": {"DOI": {"DOI": "doi:10.16904/envidat.lwf.34"}}, "meta": {}}
    )

    responses.add(
        responses.GET,
        "https://citation.doi.org/format?doi=doi:10.16904/envidat.lwf.34&style=apa&lang=fr-FR",
        body="Meusburger, K., Graf Pannatier, E., & Schaub, M. (2019). 10-HS Pfynwald (Version 2019) [Jeu de données]. EnviDat. https://doi.org/10.16904/ENVIDAT.LWF.34",
        status=200,
    )

    citation = collection.citation(format="apa", language="fr-FR")

    assert (
        citation
        == "Meusburger, K., Graf Pannatier, E., & Schaub, M. (2019). 10-HS Pfynwald (Version 2019) [Jeu de données]. EnviDat. https://doi.org/10.16904/ENVIDAT.LWF.34"
    )


def test_get_citation_returns_none_when_doi_missing():
    collection = DataCollection({"umm": {}, "meta": {}})

    assert collection.citation(format="apa", language="en-US") is None


def test_get_citation_returns_none_when_doi_empty():
    collection = DataCollection({"umm": {"DOI": {"DOI": ""}}, "meta": {}})

    assert collection.citation(format="apa", language="en-US") is None



from __future__ import annotations

import tempfile
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import urlparse

import earthaccess

if TYPE_CHECKING:
    import xarray as xr


def open_virtual_mfdataset(
    granules: list[earthaccess.DataGranule],
    group: str | None = None,
    access: str = "indirect",
    preprocess: callable | None = None,  # type: ignore
    parallel: Literal["dask", "lithops", False] = "dask",
    load: bool = True,
    reference_dir: str | None = None,
    reference_format: Literal["json", "parquet"] = "json",
    **xr_combine_nested_kwargs: Any,
) -> xr.Dataset:
    """Open multiple granules as a single virtual xarray Dataset.

    Uses NASA DMR++ metadata files to create a virtual xarray dataset with ManifestArrays. This virtual dataset can be used to create zarr reference files. See [https://virtualizarr.readthedocs.io](https://virtualizarr.readthedocs.io) for more information on virtual xarray datasets.

    > WARNING: This feature is current experimental and may change in the future. This feature relies on DMR++ metadata files which may not always be present for your dataset and you may get a `FileNotFoundError`.

    Parameters:
        granules:
            The granules to open
        group:
            Path to the netCDF4 group in the given file to open. If None, the root group will be opened. If the DMR++ file does not have groups, this parameter is ignored.
        access:
            The access method to use. One of "direct" or "indirect". Use direct when running on AWS, use indirect when running on a local machine.
        preprocess:
            A function to apply to each virtual dataset before combining
        parallel:
            Open the virtual datasets in parallel (using dask.delayed or lithops)
        load:
            If load is True, earthaccess will serialize the virtual references in order to use lazy indexing on the resulting xarray virtual ds.
        reference_dir:
            Directory to store kerchunk references. If None, a temporary directory will be created and deleted after use.
        reference_format:
            When load is True, earthaccess will serialize the references using this format, json (default) or parquet.
        xr_combine_nested_kwargs:
            Xarray arguments describing how to concatenate the datasets. Keyword arguments for xarray.combine_nested.
            See [https://docs.xarray.dev/en/stable/generated/xarray.combine_nested.html](https://docs.xarray.dev/en/stable/generated/xarray.combine_nested.html)

    Returns:
        Concatenated xarray.Dataset

    Examples:
        ```python
        >>> results = earthaccess.search_data(count=5, temporal=("2024"), short_name="MUR-JPL-L4-GLOB-v4.1")
        >>> vds = earthaccess.open_virtual_mfdataset(results, access="indirect", load=False, concat_dim="time", coords="minimal", compat="override", combine_attrs="drop_conflicts")
        >>> vds
        <xarray.Dataset> Size: 29GB
        Dimensions:           (time: 5, lat: 17999, lon: 36000)
        Coordinates:
            time              (time) int32 20B ManifestArray<shape=(5,), dtype=int32,...
            lat               (lat) float32 72kB ManifestArray<shape=(17999,), dtype=...
            lon               (lon) float32 144kB ManifestArray<shape=(36000,), dtype...
        Data variables:
            mask              (time, lat, lon) int8 3GB ManifestArray<shape=(5, 17999...
            sea_ice_fraction  (time, lat, lon) int8 3GB ManifestArray<shape=(5, 17999...
            dt_1km_data       (time, lat, lon) int8 3GB ManifestArray<shape=(5, 17999...
            analysed_sst      (time, lat, lon) int16 6GB ManifestArray<shape=(5, 1799...
            analysis_error    (time, lat, lon) int16 6GB ManifestArray<shape=(5, 1799...
            sst_anomaly       (time, lat, lon) int16 6GB ManifestArray<shape=(5, 1799...
        Attributes: (12/42)
            Conventions:                CF-1.7
            title:                      Daily MUR SST, Final product

        >>> vds.virtualize.to_kerchunk("mur_combined.json", format="json")
        >>> vds = open_virtual_mfdataset(results, access="indirect", concat_dim="time", coords='minimal', compat='override', combine_attrs="drop_conflicts")
        >>> vds
        <xarray.Dataset> Size: 143GB
        Dimensions:           (time: 5, lat: 17999, lon: 36000)
        Coordinates:
        * lat               (lat) float32 72kB -89.99 -89.98 -89.97 ... 89.98 89.99
        * lon               (lon) float32 144kB -180.0 -180.0 -180.0 ... 180.0 180.0
        * time              (time) datetime64[ns] 40B 2024-01-01T09:00:00 ... 2024-...
        Data variables:
            analysed_sst      (time, lat, lon) float64 26GB dask.array<chunksize=(1, 3600, 7200), meta=np.ndarray>
            analysis_error    (time, lat, lon) float64 26GB dask.array<chunksize=(1, 3600, 7200), meta=np.ndarray>
            dt_1km_data       (time, lat, lon) timedelta64[ns] 26GB dask.array<chunksize=(1, 4500, 9000), meta=np.ndarray>
            mask              (time, lat, lon) float32 13GB dask.array<chunksize=(1, 4500, 9000), meta=np.ndarray>
            sea_ice_fraction  (time, lat, lon) float64 26GB dask.array<chunksize=(1, 4500, 9000), meta=np.ndarray>
            sst_anomaly       (time, lat, lon) float64 26GB dask.array<chunksize=(1, 3600, 7200), meta=np.ndarray>
        Attributes: (12/42)
            Conventions:                CF-1.7
            title:                      Daily MUR SST, Final product
        ```
    """
    try:
        import virtualizarr as vz
        import xarray as xr
        from obstore.auth.earthdata import NasaEarthdataCredentialProvider
        from obstore.store import HTTPStore, S3Store
        from virtualizarr.parsers import DMRPPParser
        from virtualizarr.registry import ObjectStoreRegistry
    except ImportError as e:
        raise ImportError(
            "`earthaccess.open_virtual_dataset` requires `pip install earthaccess[virtualizarr]`"
        ) from e

    if len(granules) == 0:
        raise ValueError("No granules provided. At least one granule is required.")

    parsed_url = urlparse(granules[0].data_links(access=access)[0])
    fs = earthaccess.get_fsspec_https_session()
    if len(granules):
        collection_id = granules[0]["meta"]["collection-concept-id"]

    if access == "direct":
        credentials_endpoint, region = get_granule_credentials_endpoint_and_region(
            granules[0]
        )
        bucket = parsed_url.netloc

        if load:
            fs = earthaccess.get_s3_filesystem(endpoint=credentials_endpoint)

        s3_store = S3Store(
            bucket=bucket,
            region=region,
            credential_provider=NasaEarthdataCredentialProvider(credentials_endpoint),
            virtual_hosted_style_request=False,
            client_options={"allow_http": True},
        )
        obstore_registry = ObjectStoreRegistry({f"s3://{bucket}": s3_store})
    else:
        domain = parsed_url.netloc
        http_store = HTTPStore.from_url(
            f"https://{domain}",
            client_options={
                "default_headers": {
                    "Authorization": f"Bearer {earthaccess.__auth__.token['access_token']}",
                },
            },
        )
        obstore_registry = ObjectStoreRegistry({f"https://{domain}": http_store})

    granule_dmrpp_urls = [
        granule.data_links(access=access)[0] + ".dmrpp" for granule in granules
    ]

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="Numcodecs codecs*",
            category=UserWarning,
        )
        vmfdataset = vz.open_virtual_mfdataset(
            urls=granule_dmrpp_urls,
            registry=obstore_registry,
            parser=DMRPPParser(group=group),
            preprocess=preprocess,
            parallel=parallel,
            combine="nested",
            **xr_combine_nested_kwargs,
        )

    if load:
        if reference_dir is None:
            ref_dir = Path(tempfile.gettempdir())
        else:
            ref_dir = Path(reference_dir)
            ref_dir.mkdir(exist_ok=True, parents=True)  # type: ignore

        if group is None or group == "/":
            group_name = "root"
        else:
            group_name = group.replace("/", "_").replace(" ", "_").lstrip("_")

        ref_ = ref_dir / Path(f"{collection_id}-{group_name}.{reference_format}")
        # We still need the round trip because: https://github.com/zarr-developers/VirtualiZarr/issues/360
        vmfdataset.virtualize.to_kerchunk(str(ref_), format=reference_format)

        storage_options = {
            "remote_protocol": "s3" if access == "direct" else "https",
            "remote_options": fs.storage_options,
        }
        vds = xr.open_dataset(
            str(ref_),
            engine="kerchunk",
            storage_options=storage_options,
        )
        return vds

    return vmfdataset


def open_virtual_dataset(
    granule: earthaccess.DataGranule,
    load: bool = True,
    group: str | None = None,
    access: str = "indirect",
) -> xr.Dataset:
    """Open a granule as a single virtual xarray Dataset.

    Uses NASA DMR++ metadata files to create a virtual xarray dataset with ManifestArrays. This virtual dataset can be used to create zarr reference files. See [https://virtualizarr.readthedocs.io](https://virtualizarr.readthedocs.io) for more information on virtual xarray datasets.

    > WARNING: This feature is current experimental and may change in the future. This feature relies on DMR++ metadata files which may not always be present for your dataset and you may get a `FileNotFoundError`.

    Parameters:
        granule:
            The granule to open
        group:
            Path to the netCDF4 group in the given file to open. If None, the root group will be opened. If the DMR++ file does not have groups, this parameter is ignored.
        access:
            The access method to use. One of "direct" or "indirect". Use direct when running on AWS, use indirect when running on a local machine.

    Returns:
        xarray.Dataset

    Examples:
        ```python
        >>> results = earthaccess.search_data(count=2, temporal=("2023"), short_name="SWOT_L2_LR_SSH_Expert_2.0")
        >>> vds =  earthaccess.open_virtual_dataset(results[0], access="indirect")
        >>> vds
        <xarray.Dataset> Size: 149MB
        Dimensions:                                (num_lines: 9866, num_pixels: 69,
                                                    num_sides: 2)
        Coordinates:
            longitude                              (num_lines, num_pixels) int32 3MB ...
            latitude                               (num_lines, num_pixels) int32 3MB ...
            latitude_nadir                         (num_lines) int32 39kB ManifestArr...
            longitude_nadir                        (num_lines) int32 39kB ManifestArr...
        Dimensions without coordinates: num_lines, num_pixels, num_sides
        Data variables: (12/98)
            height_cor_xover_qual                  (num_lines, num_pixels) uint8 681kB ManifestArray<shape=(9866, 69), dtype=uint8, chunks=(9866, 69...
        >>> vds.virtualize.to_kerchunk("swot_2023_ref.json", format="json")
        ```
    """
    return open_virtual_mfdataset(
        granules=[granule],
        load=load,
        group=group,
        access=access,
        parallel=False,
        preprocess=None,
    )


def get_granule_credentials_endpoint_and_region(
    granule: earthaccess.DataGranule,
) -> tuple[str, str]:
    """Retrieve credentials endpoint for direct access granule link.

    Parameters:
        granule:
            The first granule being included in the virtual dataset.

    Returns:
        credentials_endpoint:
            The S3 credentials endpoint. If this information is in the UMM-G record, then it is used from there. If not, a query for the collection is performed and the information is taken from the UMM-C record.
        region:
            Region for the data. Defaults to us-west-2. If the credentials endpoint is retrieved from the UMM-C record for the collection, the Region information is also used from UMM-C.

    """
    credentials_endpoint = granule.get_s3_credentials_endpoint()
    region = "us-west-2"

    if credentials_endpoint is None:
        collection_results = earthaccess.search_datasets(
            count=1,
            concept_id=granule["meta"]["collection-concept-id"],
        )
        collection_s3_bucket = collection_results[0].s3_bucket()
        credentials_endpoint = collection_s3_bucket.get("S3CredentialsAPIEndpoint")
        region = collection_s3_bucket.get("Region", "us-west-2")

    if credentials_endpoint is None:
        raise ValueError("The collection did not provide an S3CredentialsAPIEndpoint")

    return credentials_endpoint, region



# package imports
import datetime as dt

import pytest
from earthaccess.search import DataCollections

valid_single_dates = [
    ("2001-12-12", "2001-12-21", "2001-12-12T00:00:00Z,2001-12-21T23:59:59Z"),
    ("2021-02-01", "", "2021-02-01T00:00:00Z,"),
    ("1999-02-01 06:00", "2009-01-01", "1999-02-01T06:00:00Z,2009-01-01T23:59:59Z"),
    (
        dt.datetime(2021, 2, 1),
        dt.datetime(2021, 2, 2),
        "2021-02-01T00:00:00Z,2021-02-02T00:00:00Z",
    ),
    (
        "2019-03-10T00:00:00Z",
        "2019-03-11T23:59:59Z",
        "2019-03-10T00:00:00Z,2019-03-11T23:59:59Z",
    ),
    (
        "2019-03-10T00:00:00Z",
        "2019-03-10T00:00:00-01:00",
        "2019-03-10T00:00:00Z,2019-03-10T01:00:00Z",
    ),
]

invalid_single_dates = [
    ("2001-12-45", "2001-12-21", None),
    ("2021w1", "", None),
    ("2999-02-01", "2009-01-01", None),
]


def test_no_default_params():
    query = DataCollections()
    assert len(query.params) == 0


def test_query_can_find_cloud_provider():
    query = DataCollections().daac("PODAAC").cloud_hosted(True)
    assert query.params["provider"] == "POCLOUD"
    query = DataCollections().cloud_hosted(True).daac("PODAAC")
    assert query.params["provider"] == "POCLOUD"
    # SEDAC does not have a cloud provider, so it should default to the on prem provider
    query = DataCollections().cloud_hosted(True).daac("SEDAC")
    assert query.params["provider"] == "ESDIS"
    query = DataCollections().daac("ASDC").cloud_hosted(True)
    assert query.params["provider"] == "LARC_CLOUD"
    query = DataCollections().cloud_hosted(True).daac("ASDC")
    assert query.params["provider"] == "LARC_CLOUD"


def test_querybuilder_can_handle_doi():
    doi = "10.5067/AQR50-3Q7CS"
    query = DataCollections().doi(doi)
    assert query.params["doi"] == doi
    query = DataCollections().cloud_hosted(True).daac("PODAAC").doi(doi)
    assert query.params["doi"] == doi


def test_querybuilder_can_handle_has_granules():
    query = DataCollections().has_granules(False)
    assert not query.params["has_granules"]
    query = DataCollections().has_granules(True)
    assert query.params["has_granules"]
    query = DataCollections().has_granules(None)
    assert "has_granules" not in query.params


@pytest.mark.parametrize("start,end,expected", valid_single_dates)
def test_query_can_parse_single_dates(start, end, expected):
    query = DataCollections().temporal(start, end)
    assert query.params["temporal"][0] == expected


@pytest.mark.parametrize("start,end,expected", invalid_single_dates)
def test_query_can_handle_invalid_dates(start, end, expected):
    query = DataCollections()
    try:
        query = query.temporal(start, end)
    except Exception as e:
        assert isinstance(e, ValueError)
        assert "temporal" not in query.params



from typing import Any, List, Optional

import requests

from cmr import ServiceQuery

from .auth import Auth
from .utils import _search as search


class DataServices(ServiceQuery):
    """A Service client for NASA CMR that returns data on collection services.

    API: https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html#service
    """

    _format = "umm_json"

    def __init__(self, auth: Optional[Auth] = None, *args: Any, **kwargs: Any) -> None:
        """Build an instance of DataService to query CMR.

        auth is an optional parameter for queries that need authentication,
        e.g. restricted datasets.

        Parameters:
            auth: An authenticated `Auth` instance.
        """
        super().__init__(*args, **kwargs)
        self._debug = False

        self.session = (
            auth.get_session()
            if auth is not None and auth.authenticated
            else requests.sessions.Session()
        )

    def get(self, limit: int = 2000) -> List:
        """Get all service results up to some limit.

        Parameters
            limit (int): The number of results to return

        Returns:
            Query results as a list
        """
        return search.get_results(self.session, self, limit)



import datetime as dt
import logging
from inspect import getmembers, ismethod

import requests
from typing_extensions import (
    Any,
    Iterable,
    List,
    Optional,
    Self,
    Sequence,
    SupportsFloat,
    Tuple,
    TypeAlias,
    Union,
    override,
)

from cmr import CollectionQuery, GranuleQuery

from .auth import Auth
from .daac import find_provider, find_provider_by_shortname
from .results import DataCollection, DataGranule
from .utils._search import get_results

logger = logging.getLogger(__name__)

FloatLike: TypeAlias = Union[str, SupportsFloat]
PointLike: TypeAlias = Tuple[FloatLike, FloatLike]


class DataCollections(CollectionQuery):
    """Placeholder.

    ???+ Info
        The DataCollection class queries against
        https://cmr.earthdata.nasa.gov/search/collections.umm_json,
        the response has to be in umm_json to use the result classes.
    """

    _fields: Optional[List[str]] = None
    _format = "umm_json"

    def __init__(self, auth: Optional[Auth] = None, *args: Any, **kwargs: Any) -> None:
        """Builds an instance of DataCollections to query the CMR.

        Parameters:
            auth: An authenticated `Auth` instance. This is an optional parameter
                for queries that need authentication, e.g. restricted datasets.
        """
        super().__init__(*args, **kwargs)

        self.session = (
            auth.get_session()
            if auth and auth.authenticated
            else requests.sessions.Session()
        )

        if auth:
            self.mode(auth.system.cmr_base_url)

        self._debug = False

    @override
    def hits(self) -> int:
        """Returns the number of hits the current query will return.

        This is done by making a lightweight query to CMR and inspecting the returned headers.
        Restricted datasets will always return zero results even if there are results.

        Returns:
            The number of results reported by the CMR.

        Raises:
            RuntimeError: The CMR query failed.
        """
        url = self._build_url()

        response = self.session.get(url, headers=self.headers, params={"page_size": 0})

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            raise RuntimeError(ex.response.text) from ex

        return int(response.headers["CMR-Hits"])

    @override
    def get(self, limit: int = 2000) -> List[DataCollection]:
        """Get all the collections (datasets) that match with our current parameters
        up to some limit, even if spanning multiple pages.

        ???+ Tip
            The default page size is 2000, we need to be careful with the request size
            because all the JSON elements will be loaded into memory. This is more of an
            issue with granules than collections as they can be potentially millions of
            them.

        Parameters:
            limit: The number of results to return

        Returns:
            Query results as a (possibly empty) list of `DataCollection` instances.

        Raises:
            RuntimeError: The CMR query failed.
        """
        return [
            DataCollection(collection, self._fields)
            for collection in get_results(self.session, self, limit)
        ]

    @override
    def concept_id(self, IDs: Sequence[str]) -> Self:
        """Filter by concept ID.

        For example: C1299783579-LPDAAC_ECS or G1327299284-LPDAAC_ECS,
        S12345678-LPDAAC_ECS

        Collections, granules, tools, services are uniquely identified with this ID.

        * If providing a collection's concept ID, it will filter by granules associated
          with that collection.
        * If providing a granule's concept ID, it will uniquely identify those granules.
        * If providing a tool's concept ID, it will uniquely identify those tools.
        * If providing a service's concept ID, it will uniquely identify those services.

        Parameters:
            IDs: ID(s) to search by. Can be provided as a string or list of strings.

        Returns:
            self

        Raises:
            ValueError: An ID does not start with a valid prefix.
        """
        return super().concept_id(IDs)

    @override
    def keyword(self, text: str) -> Self:
        """Case-insensitive and wildcard (*) search through over two dozen fields in
        a CMR collection record. This allows for searching against fields like
        summary and science keywords.

        Parameters:
            text: text to search for

        Returns:
            self
        """
        return super().keyword(text)

    def doi(self, doi: str) -> Self:
        """Search datasets by DOI.

        ???+ Tip
            Not all datasets have an associated DOI, also DOI search works
            only at the dataset level but not the granule (data) level.
            We need to search by DOI, grab the concept_id and then get the data.

        Parameters:
            doi: DOI of a datasets, e.g. 10.5067/AQR50-3Q7CS

        Returns:
            self

        Raises:
            TypeError: `doi` is not of type `str`.
        """
        if not isinstance(doi, str):
            raise TypeError("doi must be of type str")

        self.params["doi"] = doi
        return self

    def instrument(self, instrument: str) -> Self:
        """Search datasets by instrument.

        ???+ Tip
            Not all datasets have an associated instrument. This works
            only at the dataset level but not the granule (data) level.

        Parameters:
            instrument (String): instrument of a datasets, e.g. instrument=GEDI

        Returns:
            self

        Raises:
            TypeError: `instrument` is not of type `str`.
        """
        if not isinstance(instrument, str):
            raise TypeError("instrument must be of type str")

        self.params["instrument"] = instrument
        return self

    def project(self, project: str) -> Self:
        """Search datasets by associated project.

        ???+ Tip
            Not all datasets have an associated project. This works
            only at the dataset level but not the granule (data) level.
            Will return datasets across DAACs matching the project.

        Parameters:
            project (String): associated project of a datasets, e.g. project=EMIT

        Returns:
            self

        Raises:
            TypeError: `project` is not of type `str`.
        """
        if not isinstance(project, str):
            raise TypeError("project must be of type str")

        self.params["project"] = project
        return self

    @override
    def parameters(self, **kwargs: Any) -> Self:
        """Provide query parameters as keyword arguments. The keyword needs to match the name
        of the method, and the value should either be the value or a tuple of values.

        ???+ Example
            ```python
            query = DataCollections.parameters(
                short_name="AST_L1T",
                temporal=("2015-01","2015-02"),
                point=(42.5, -101.25)
            )
            ```

        Returns:
            self

        Raises:
            ValueError: The name of a keyword argument is not the name of a method.
            TypeError: The value of a keyword argument is not an argument or tuple
                of arguments matching the number and type(s) of the method's parameters.
        """
        methods = dict(getmembers(self, predicate=ismethod))

        for key, val in kwargs.items():
            # verify the key matches one of our methods
            if key not in methods:
                raise ValueError("Unknown key {}".format(key))

            # call the method
            if isinstance(val, tuple):
                methods[key](*val)
            else:
                methods[key](val)

        return self

    def print_help(self, method: str = "fields") -> None:
        """Prints the help information for a given method."""
        print("Class components: \n")  # noqa: T201
        print(  # noqa: T201
            [method for method in dir(self) if method.startswith("_") is False]
        )
        help(getattr(self, method))

    def fields(self, fields: Optional[List[str]] = None) -> Self:
        """Masks the response by only showing the fields included in this list.

        Parameters:
            fields (List): list of fields to show. These fields come from the UMM model
                (e.g. Abstract, Title).

        Returns:
            self
        """
        self._fields = fields
        return self

    def debug(self, debug: bool = True) -> Self:
        """If True, prints the actual query to CMR. Note that the pagination happens in
        the headers.

        Parameters:
            debug (Boolean): If `True`, print the CMR query.

        Returns:
            self
        """
        self._debug = debug
        return self

    def has_granules(self, has_granules: bool | None = True) -> Self:
        """Match only collections with granules, without granules, or either.

        Parameters:
            has_granules:
                If `True`, only return collections with granules. If
                `False`, only return collections without granules.
                If `None`, return both types of collections.

        Returns:
            self
        """
        if has_granules is not None and not isinstance(has_granules, bool):
            raise TypeError("has_granules must be of type bool or None")

        if has_granules is not None:
            self.params["has_granules"] = has_granules
        elif "has_granules" in self.params:
            del self.params["has_granules"]

        return self

    def cloud_hosted(self, cloud_hosted: bool = True) -> Self:
        """Only match granules that are hosted in the cloud. This is valid for public
        collections.

        ???+ Tip
            Cloud hosted collections can be public or restricted.
            Restricted collections will not be matched using this parameter

        Parameters:
            cloud_hosted: If `True`, obtain only cloud-hosted collections.

        Returns:
            self

        Raises:
            TypeError: `cloud_hosted` is not of type `bool`.
        """
        if not isinstance(cloud_hosted, bool):
            raise TypeError("cloud_hosted must be of type bool")

        self.params["cloud_hosted"] = cloud_hosted
        if hasattr(self, "DAAC"):
            provider = find_provider(self.DAAC, cloud_hosted)
            self.params["provider"] = provider
        return self

    @override
    def provider(self, provider: str) -> Self:
        """Only match collections from a given provider.

        A NASA datacenter or DAAC can have one or more providers.
        E.g., PODAAC is a data center or DAAC; PODAAC is the default provider for
        on-premises data, POCLOUD is the PODAAC provider for their data in the cloud.

        Parameters:
            provider: a provider code for any DAAC, e.g. POCLOUD, NSIDC_CPRD, etc.

        Returns:
            self
        """
        self.params["provider"] = provider
        return self

    def data_center(self, data_center_name: str) -> Self:
        """An alias for the `daac` method.

        Parameters:
            data_center_name: DAAC shortname, e.g. NSIDC, PODAAC, GES_DISC

        Returns:
            self
        """
        return self.daac(data_center_name)

    def daac(self, daac_short_name: str) -> Self:
        """Only match collections for a given DAAC, by default the on-prem collections
        for the DAAC.

        Parameters:
            daac_short_name: a DAAC shortname, e.g. NSIDC, PODAAC, GES_DISC

        Returns:
            self
        """
        if "cloud_hosted" in self.params:
            cloud_hosted = self.params["cloud_hosted"]
        else:
            cloud_hosted = False
        self.DAAC = daac_short_name
        self.params["provider"] = find_provider(daac_short_name, cloud_hosted)
        return self

    @override
    def temporal(
        self,
        date_from: Optional[Union[str, dt.date, dt.datetime]] = None,
        date_to: Optional[Union[str, dt.date, dt.datetime]] = None,
        exclude_boundary: bool = False,
    ) -> Self:
        """Filter by an open or closed date range. Dates can be provided as date objects
        or ISO 8601 strings. Multiple ranges can be provided by successive method calls.

        ???+ Tip
            Giving either `datetime.date(YYYY, MM, DD)` or `"YYYY-MM-DD"` as the `date_to`
            parameter includes that entire day (i.e. the time is set to `23:59:59`).
            Using `datetime.datetime(YYYY, MM, DD)` is different, because `datetime.datetime`
            objects have `00:00:00` as their built-in default.

        Parameters:
            date_from: start of temporal range
            date_to: end of temporal range
            exclude_boundary: whether or not to exclude the date_from/to in
                the matched range.

        Returns:
            self

        Raises:
            ValueError: `date_from` or `date_to` is a non-`None` value that is
                neither a datetime object nor a string that can be parsed as a datetime
                object; or `date_from` and `date_to` are both datetime objects (or
                parsable as such) and `date_from` is after `date_to`.
        """
        return super().temporal(date_from, date_to, exclude_boundary)


class DataGranules(GranuleQuery):
    """A Granule oriented client for NASA CMR.

    API: https://cmr.earthdata.nasa.gov/search/site/docs/search/api.html
    """

    _format = "umm_json"

    def __init__(self, auth: Optional[Auth] = None, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.session = (
            auth.get_session()
            if auth and auth.authenticated
            else requests.sessions.Session()
        )

        if auth:
            self.mode(auth.system.cmr_base_url)

        self._debug = False

    @override
    def hits(self) -> int:
        """Returns the number of hits the current query will return.

        This is done by making a lightweight query to CMR and inspecting the returned
        headers.

        Returns:
            The number of results reported by the CMR.

        Raises:
            RuntimeError: The CMR query failed.
        """
        url = self._build_url()

        response = self.session.get(url, headers=self.headers, params={"page_size": 0})

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            if ex.response is not None:
                raise RuntimeError(ex.response.text) from ex
            else:
                raise RuntimeError(str(ex)) from ex

        return int(response.headers["CMR-Hits"])

    @override
    def get(self, limit: int = 2000) -> List[DataGranule]:
        """Get all the collections (datasets) that match with our current parameters
        up to some limit, even if spanning multiple pages.

        ???+ Tip
            The default page size is 2000, we need to be careful with the request size
            because all the JSON elements will be loaded into memory. This is more of an
            issue with granules than collections as they can be potentially millions of
            them.

        Parameters:
            limit: The number of results to return.

        Returns:
            Query results as a (possibly empty) list of `DataGranules` instances.

        Raises:
            RuntimeError: The CMR query failed.
        """
        response = get_results(self.session, self, limit)
        cloud = len(response) > 0 and self._is_cloud_hosted(response[0])

        return [DataGranule(granule, cloud_hosted=cloud) for granule in response]

    @override
    def parameters(self, **kwargs: Any) -> Self:
        """Provide query parameters as keyword arguments. The keyword needs to match the
        name of the method, and the value should either be the value or a tuple of
        values.

        ???+ Example
            ```python
            query = DataCollections.parameters(
                short_name="AST_L1T",
                temporal=("2015-01","2015-02"),
                point=(42.5, -101.25)
            )
            ```

        Returns:
            self

        Raises:
            ValueError: The name of a keyword argument is not the name of a method.
            TypeError: The value of a keyword argument is not an argument or tuple
                of arguments matching the number and type(s) of the method's parameters.
        """
        methods = {}
        for name, func in getmembers(self, predicate=ismethod):
            methods[name] = func

        for key, val in kwargs.items():
            # verify the key matches one of our methods
            if key not in methods:
                raise ValueError("Unknown key {}".format(key))

            # call the method
            if isinstance(val, tuple):
                methods[key](*val)
            else:
                methods[key](val)

        return self

    @override
    def provider(self, provider: str) -> Self:
        """Only match collections from a given provider.

        A NASA datacenter or DAAC can have one or more providers.
        For example, PODAAC is a data center or DAAC,
        PODAAC is the default provider for on-prem data, and POCLOUD is
        the PODAAC provider for their data in the cloud.

        Parameters:
            provider: a provider code for any DAAC, e.g. POCLOUD, NSIDC_CPRD, etc.

        Returns:
            self
        """
        self.params["provider"] = provider
        return self

    def data_center(self, data_center_name: str) -> Self:
        """An alias for the `daac` method.

        Parameters:
            data_center_name (String): DAAC shortname, e.g. NSIDC, PODAAC, GES_DISC

        Returns:
            self
        """
        return self.daac(data_center_name)

    def daac(self, daac_short_name: str) -> Self:
        """Only match collections for a given DAAC. Default to on-prem collections for
        the DAAC.

        Parameters:
            daac_short_name: a DAAC shortname, e.g. NSIDC, PODAAC, GES_DISC

        Returns:
            self
        """
        if "cloud_hosted" in self.params:
            cloud_hosted = self.params["cloud_hosted"]
        else:
            cloud_hosted = False
        self.DAAC = daac_short_name
        self.params["provider"] = find_provider(daac_short_name, cloud_hosted)
        return self

    @override
    def orbit_number(
        self,
        orbit1: FloatLike,
        orbit2: Optional[FloatLike] = None,
    ) -> Self:
        """Filter by the orbit number the granule was acquired during. Either a single
        orbit can be targeted or a range of orbits.

        Parameter:
            orbit1: orbit to target (lower limit of range when orbit2 is provided)
            orbit2: upper limit of range

        Returns:
            self
        """
        return super().orbit_number(orbit1, orbit2)

    def cloud_hosted(self, cloud_hosted: bool = True) -> Self:
        """Only match granules that are hosted in the cloud.
        This is valid for public collections and when using the short_name parameter.
        Concept-Id is unambiguous.

        ???+ Tip
            Cloud-hosted collections can be public or restricted.
            Restricted collections will not be matched using this parameter.

        Parameters:
            cloud_hosted: If `True`, obtain only granules from cloud-hosted collections.

        Returns:
            self

        Raises:
            TypeError: `cloud_hosted` is not of type `bool`.
        """
        if not isinstance(cloud_hosted, bool):
            raise TypeError("cloud_hosted must be of type bool")

        if "short_name" in self.params:
            provider = find_provider_by_shortname(
                self.params["short_name"], cloud_hosted
            )
            if provider is not None:
                self.params["provider"] = provider
        return self

    def granule_name(self, granule_name: str | Iterable[str]) -> Self:
        """Find granules matching either granule ur or producer granule id,
        queries using the readable_granule_name metadata field.

        ???+ Tip
            We can use wildcards on a granule name to further refine our search,
            e.g. `MODGRNLD.*.daily.*`.

        Parameters:
            granule_name: granule name (accepts wildcards)

        Returns:
            self

        Raises:
            TypeError: if `granule_name` is not of type `str` or `Iterable[str]`.
        """
        if not isinstance(granule_name, Iterable):
            raise TypeError(
                "granule_name must be of type string or Iterable of strings"
            )
        if not isinstance(granule_name, str):
            # Convert iterable to list of strings. Since str is also Iterable, make
            # sure we don't do this when granule_name is a string, otherwise
            # we would get a list of individual characters.
            granule_name = [str(name) for name in granule_name]

        self.params["readable_granule_name"] = granule_name
        self.params["options[readable_granule_name][pattern]"] = True
        return self

    @override
    def online_only(self, online_only: bool = True) -> Self:
        """Only match granules that are listed online and not available for download.
        The inverse of this method is `downloadable`.

        Parameters:
            online_only: If `True`, obtain only granules that are online (not
                downloadable)

        Returns:
            self

        Raises:
            TypeError: `online_only` is not of type `bool`.
        """
        return super().online_only(online_only)

    @override
    def day_night_flag(self, day_night_flag: str) -> Self:
        """Filter by period of the day the granule was collected during.

        Parameters:
            day_night_flag: "day", "night", or "unspecified"

        Returns:
            self

        Raises:
            TypeError: `day_night_flag` is not of type `str`.
            ValueError: `day_night_flag` is not one of `"day"`, `"night"`, or
                `"unspecified"`.
        """
        return super().day_night_flag(day_night_flag)

    @override
    def instrument(self, instrument: str) -> Self:
        """Filter by the instrument associated with the granule.

        Parameters:
            instrument: name of the instrument

        Returns:
            self

        Raises:
            ValueError: `instrument` is not a non-empty string.
        """
        return super().instrument(instrument)

    @override
    def platform(self, platform: str) -> Self:
        """Filter by the satellite platform the granule came from.

        Parameters:
            platform: name of the satellite

        Returns:
            self

        Raises:
            ValueError: `platform` is not a non-empty string.
        """
        return super().platform(platform)

    @override
    def cloud_cover(
        self,
        min_cover: Optional[FloatLike] = 0,
        max_cover: Optional[FloatLike] = 100,
    ) -> Self:
        """Filter by the percentage of cloud cover present in the granule.

        Parameters:
            min_cover: minimum percentage of cloud cover
            max_cover: maximum percentage of cloud cover

        Returns:
            self

        Raises:
            ValueError: `min_cover` or `max_cover` is not convertible to a float,
                or `min_cover` is greater than `max_cover`.
        """
        return super().cloud_cover(min_cover, max_cover)

    def _valid_state(self) -> bool:
        # spatial params must be paired with a collection limiting parameter
        spatial_keys = ["point", "polygon", "bounding_box", "line"]
        collection_keys = ["short_name", "entry_title", "concept_id"]

        if any(key in self.params for key in spatial_keys):
            if not any(key in self.params for key in collection_keys):
                return False

        # all good then
        return True

    def _is_cloud_hosted(self, granule: Any) -> bool:
        """Check if a granule record, from CMR, advertises "direct access"."""
        if "RelatedUrls" not in granule["umm"]:
            return False

        direct_def = "GET DATA VIA DIRECT ACCESS"
        for link in granule["umm"]["RelatedUrls"]:
            if "protected" in link["URL"] or link["Type"] == direct_def:
                return True
        return False

    @override
    def short_name(self, short_name: str) -> Self:
        """Filter by short name (aka product or collection name).

        Parameters:
            short_name: name of a collection

        Returns:
            self
        """
        return super().short_name(short_name)

    def debug(self, debug: bool = True) -> Self:
        """If True, prints the actual query to CMR, notice that the pagination happens
        in the headers.

        Parameters:
            debug: If `True`, print the CMR query.

        Returns:
            self
        """
        self._debug = debug
        return self

    @override
    def temporal(
        self,
        date_from: Optional[Union[str, dt.date, dt.datetime]] = None,
        date_to: Optional[Union[str, dt.date, dt.datetime]] = None,
        exclude_boundary: bool = False,
    ) -> Self:
        """Filter by an open or closed date range. Dates can be provided as date objects
        or ISO 8601 strings. Multiple ranges can be provided by successive method calls.

        ???+ Tip
            Giving either `datetime.date(YYYY, MM, DD)` or `"YYYY-MM-DD"` as the `date_to`
            parameter includes that entire day (i.e. the time is set to `23:59:59`).
            Using `datetime.datetime(YYYY, MM, DD)` is different, because `datetime.datetime`
            objects have `00:00:00` as their built-in default.

        Parameters:
            date_from: start of temporal range
            date_to: end of temporal range
            exclude_boundary: whether to exclude the date_from and date_to in the matched range

        Returns:
            self

        Raises:
            ValueError: `date_from` or `date_to` is a non-`None` value that is
                neither a datetime object nor a string that can be parsed as a datetime
                object; or `date_from` and `date_to` are both datetime objects (or
                parsable as such) and `date_from` is after `date_to`.
        """
        return super().temporal(date_from, date_to, exclude_boundary)

    @override
    def version(self, version: str) -> Self:
        """Filter by version. Note that CMR defines this as a string. For example,
        MODIS version 6 products must be searched for with "006".

        Parameters:
            version: version string

        Returns:
            self
        """
        return super().version(version)

    @override
    def point(self, lon: FloatLike, lat: FloatLike) -> Self:
        """Filter by granules that include a geographic point.

        Parameters:
            lon: longitude of geographic point
            lat: latitude of geographic point

        Returns:
            self

        Raises:
            ValueError: `lon` or `lat` cannot be converted to a float.
        """
        return super().point(lon, lat)

    @override
    def polygon(self, coordinates: Sequence[PointLike]) -> Self:
        """Filter by granules that overlap a polygonal area. Must be used in combination
        with a collection filtering parameter such as short_name or entry_title.

        Parameters:
            coordinates: list of (lon, lat) tuples

        Returns:
            self

        Raises:
            ValueError: `coordinates` is not a sequence of at least 4 coordinate
                pairs, any of the coordinates cannot be converted to a float, or the
                first and last coordinate pairs are not equal.
        """
        return super().polygon(coordinates)

    @override
    def bounding_box(
        self,
        lower_left_lon: FloatLike,
        lower_left_lat: FloatLike,
        upper_right_lon: FloatLike,
        upper_right_lat: FloatLike,
    ) -> Self:
        """Filter by granules that overlap a bounding box. Must be used in combination
        with a collection filtering parameter such as short_name or entry_title.

        Parameters:
            lower_left_lon: lower left longitude of the box
            lower_left_lat: lower left latitude of the box
            upper_right_lon: upper right longitude of the box
            upper_right_lat: upper right latitude of the box

        Returns:
            self

        Raises:
            ValueError: A coordinate could not be converted to a float.
        """
        return super().bounding_box(
            lower_left_lon, lower_left_lat, upper_right_lon, upper_right_lat
        )

    @override
    def line(self, coordinates: Sequence[PointLike]) -> Self:
        """Filter by granules that overlap a series of connected points. Must be used
        in combination with a collection filtering parameter such as short_name or
        entry_title.

        Parameters:
            coordinates: a list of (lon, lat) tuples

        Returns:
            self

        Raises:
            ValueError: `coordinates` is not a sequence of at least 2 coordinate
                pairs, or a coordinate could not be converted to a float.
        """
        return super().line(coordinates)

    @override
    def downloadable(self, downloadable: bool = True) -> Self:
        """Only match granules that are available for download. The inverse of this
        method is `online_only`.

        Parameters:
            downloadable: If `True`, obtain only granules that are downloadable.

        Returns:
            self

        Raises:
            TypeError: `downloadable` is not of type `bool`.
        """
        return super().downloadable(downloadable)

    def doi(self, doi: str) -> Self:
        """Search data granules by DOI.

        ???+ Tip
            Not all datasets have an associated DOI, internally if a DOI is found
            earthaccess will grab the concept_id for the query to CMR.

        Parameters:
            doi: DOI of a dataset, e.g. 10.5067/AQR50-3Q7CS

        Returns:
            self

        Raises:
            RuntimeError: The CMR query to get the collection for the DOI fails.
        """
        # TODO consider deferring this query until the search is executed
        collection = DataCollections().doi(doi).get()

        # TODO consider raising an exception when there are multiple collections, since
        # we can't know which one the user wants, and choosing one is arbitrary.
        if len(collection) > 0:
            concept_id = collection[0].concept_id()
            self.params["concept_id"] = concept_id
        else:
            # TODO consider removing this print statement since we don't print such
            # a message in other cases where no results are found.  Seems arbitrary.
            logger.info(
                f"earthaccess couldn't find any associated collections with the DOI: {doi}"
            )

        return self



"""Earthdata Environments/Systems module."""

from dataclasses import dataclass

from typing_extensions import NewType

from cmr import CMR_OPS, CMR_UAT

CMRBaseURL = NewType("CMRBaseURL", str)
EDLHostname = NewType("EDLHostname", str)
StatusURL = NewType("StatusURL", str)
StatusApiURL = NewType("StatusApiURL", str)


@dataclass(frozen=True)
class System:
    """Host URL options, for different Earthdata domains."""

    cmr_base_url: CMRBaseURL
    status_url: StatusURL
    status_api_url: StatusApiURL
    edl_hostname: EDLHostname


PROD = System(
    CMRBaseURL(CMR_OPS),
    StatusURL("https://status.earthdata.nasa.gov/"),
    StatusApiURL("https://status.earthdata.nasa.gov/api/v1/statuses"),
    EDLHostname("urs.earthdata.nasa.gov"),
)
UAT = System(
    CMRBaseURL(CMR_UAT),
    StatusURL("https://status.uat.earthdata.nasa.gov/"),
    StatusApiURL("https://status.uat.earthdata.nasa.gov/api/v1/statuses"),
    EDLHostname("uat.urs.earthdata.nasa.gov"),
)



from unittest.mock import patch

import pytest
from earthaccess.dmrpp_zarr import get_granule_credentials_endpoint_and_region
from earthaccess.results import DataCollection, DataGranule

granule_no_credentials_endpoint = DataGranule(
    {
        "meta": {
            "collection-concept-id": "C1234-PROV",
        },
        "umm": {
            "RelatedUrls": [
                {
                    "URL": "https://data.earthdata.nasa.gov/data.h5",
                    "Type": "GET DATA",
                },
            ],
        },
    },
    cloud_hosted=True,
)


@patch("earthaccess.search_datasets")
def test_get_granule_credentials_and_region_from_granule(mock_search_datasets):
    """If credentials endpoint is populated in the granule it is used."""
    granule_credentials_endpoint = (
        "https://archive.daac.earthdata.nasa.gov/s3credentials"
    )
    granule = DataGranule(
        {
            "meta": {
                "collection-concept-id": "C1234-PROV",
            },
            "umm": {
                "RelatedUrls": [
                    {
                        "URL": "https://data.earthdata.nasa.gov/data.h5",
                        "Type": "GET DATA",
                    },
                    {
                        "URL": "s3://daac-cumulus-prod-protected/data.h5",
                        "Type": "GET DATA VIA DIRECT ACCESS",
                    },
                    {
                        "URL": granule_credentials_endpoint,
                        "Type": "VIEW RELATED INFORMATION",
                    },
                ],
            },
        },
        cloud_hosted=True,
    )

    # Credentials endpoint is retrieved from the granule, the region is set to
    # a default value of us-west-2.
    assert get_granule_credentials_endpoint_and_region(granule) == (
        granule_credentials_endpoint,
        "us-west-2",
    )

    # Ensure no attempt was made to retrieve collection information
    mock_search_datasets.assert_not_called()


@patch("earthaccess.search_datasets")
def test_get_granule_credentials_and_region_from_collection(mock_search_datasets):
    """If granule does not have credentials, those for a collection are retrieved."""
    collection_credentials_endpoint = (
        "https://archive.other-daac.earthdata.nasa.gov/s3credentials"
    )
    collection_region = "us-east-1"

    mock_search_datasets.return_value = [
        DataCollection(
            {
                "meta": {
                    "concept-id": "C1234-PROV",
                },
                "umm": {
                    "DirectDistributionInformation": {
                        "Region": collection_region,
                        "S3CredentialsAPIEndpoint": collection_credentials_endpoint,
                    },
                },
            }
        ),
    ]
    # Credentials endpoint is retrieved from the collection, the region is also
    # retrieved from the collection.
    assert get_granule_credentials_endpoint_and_region(
        granule_no_credentials_endpoint
    ) == (
        collection_credentials_endpoint,
        collection_region,
    )

    # An attempt was made to retrieve collection information
    mock_search_datasets.assert_called_once_with(count=1, concept_id="C1234-PROV")


@patch("earthaccess.search_datasets")
def test_get_granule_credentials_from_collection_default_region(mock_search_datasets):
    """If a collection does not have a region, the default of us-west-2 is used."""
    collection_credentials_endpoint = (
        "https://archive.other-daac.earthdata.nasa.gov/s3credentials"
    )

    mock_search_datasets.return_value = [
        DataCollection(
            {
                "meta": {
                    "concept-id": "C1234-PROV",
                },
                "umm": {
                    "DirectDistributionInformation": {
                        "S3CredentialsAPIEndpoint": collection_credentials_endpoint,
                    },
                },
            }
        ),
    ]
    # Credentials endpoint is retrieved from the collection, the region is also
    # retrieved from the collection.
    assert get_granule_credentials_endpoint_and_region(
        granule_no_credentials_endpoint
    ) == (
        collection_credentials_endpoint,
        "us-west-2",
    )

    # An attempt was made to retrieve collection information
    mock_search_datasets.assert_called_once_with(count=1, concept_id="C1234-PROV")


@patch("earthaccess.search_datasets")
def test_get_granule_credentials_no_collection_endpoint(mock_search_datasets):
    """Exception raised if a collection does not have an S3CredentialsAPIEndpoint."""
    mock_search_datasets.return_value = [
        DataCollection(
            {
                "meta": {
                    "concept-id": "C1234-PROV",
                },
                "umm": {
                    "DirectDistributionInformation": {
                        "Region": "us-east-1",
                    },
                },
            }
        ),
    ]
    # Credentials endpoint is retrieved from the collection, the region is also
    # retrieved from the collection.

    with pytest.raises(ValueError, match="did not provide an S3CredentialsAPIEndpoint"):
        get_granule_credentials_endpoint_and_region(granule_no_credentials_endpoint)

    # An attempt was made to retrieve collection information
    mock_search_datasets.assert_called_once_with(count=1, concept_id="C1234-PROV")



import fsspec
from dask.distributed import Client
from earthaccess.store import EarthAccessFile


def test_serialization():
    fs = fsspec.filesystem("memory")
    foo = "foo"
    bar = b"bar"
    with fs.open(foo, mode="wb") as f:
        f.write(bar)
    f = fs.open(foo, mode="rb")
    earthaccess_file = EarthAccessFile(f, granule=foo)
    client = Client()
    future = client.submit(lambda f: f.read(), earthaccess_file)
    assert future.result() == bar
    # cleanup
    client.shutdown()
    fs.store.clear()



import logging
import os
from pathlib import Path
from unittest.mock import patch

import earthaccess
import pytest
import responses
from earthaccess.exceptions import ServiceOutage
from earthaccess.system import PROD, UAT

logger = logging.getLogger(__name__)


dataset_valid_params = [
    {"data_center": "NSIDC", "cloud_hosted": True},
    {"keyword": "aerosol", "cloud_hosted": False},
    {"daac": "NSIDC", "keyword": "ocean"},
]

granules_valid_params = [
    {
        "data_center": "NSIDC",
        "short_name": "ATL08",
        "cloud_hosted": True,
        # Chiapas, Mexico
        "bounding_box": (-92.86, 16.26, -91.58, 16.97),
    },
    {
        "concept_id": "C2021957295-LPCLOUD",
        "day_night_flag": "day",
        "cloud_cover": (0, 20),
        # Southern Ireland
        "bounding_box": (-10.15, 51.61, -7.59, 52.43),
    },
]

services = ("Earthdata Login", "Common Metadata Repository")
expected = {service: "Unknown" for service in services}

nasa_statuses = {
    PROD: {
        "success": True,
        "statuses": [
            {"name": "MMT (SIT)", "status": "OK"},
            {"name": "Status App Test Application", "status": "OK"},
            {"name": "NSIDC DAAC OpenAltimetry", "status": "OK"},
            {"name": "ED Pub", "status": "OK"},
            {"name": "Earthdata Login", "status": "OK"},
            {"name": "Common Metadata Repository (PROD)", "status": "OK"},
        ],
    },
    UAT: {
        "success": True,
        "statuses": [
            {"name": "MMT (SIT)", "status": "OK"},
            {"name": "Status App Test Application", "status": "OK"},
            {"name": "NSIDC DAAC OpenAltimetry", "status": "OK"},
            {"name": "ED Pub", "status": "OK"},
            {"name": "Earthdata Login (UAT)", "status": "OK"},
            {"name": "Common Metadata Repository", "status": "OK"},
        ],
    },
}


@pytest.mark.parametrize(
    "system,nasa_response,expected",
    [
        (
            PROD,
            nasa_statuses[PROD],
            {"Earthdata Login": "OK", "Common Metadata Repository": "OK"},
        ),
        (
            UAT,
            nasa_statuses[UAT],
            {"Earthdata Login": "OK", "Common Metadata Repository": "OK"},
        ),
    ],
)
@responses.activate
def test_earthdata_status_happy_path(system, nasa_response, expected):
    responses.add(
        responses.GET,
        system.status_api_url,
        json=nasa_response,
        status=200,
    )

    result = earthaccess.status(system)
    assert isinstance(result, dict)
    assert result == expected


@responses.activate
def test_earthdata_status_service_outage():
    responses.add(
        responses.GET,
        "https://status.earthdata.nasa.gov/api/v1/statuses",
        json={
            "statuses": [
                {"name": "Earthdata Login", "status": "Degraded"},
                {"name": "Common Metadata Repository", "status": "OK"},
            ]
        },
        status=200,
    )
    with pytest.raises(ServiceOutage):
        earthaccess.status(raise_on_outage=True)


@responses.activate
def test_earthdata_status_malformed_json():
    responses.add(
        responses.GET,
        "https://status.earthdata.nasa.gov/api/v1/statuses",
        body='{"x": 5,}',
        status=200,
    )

    result = earthaccess.status()
    assert isinstance(result, dict)
    assert result == expected


@responses.activate
def test_earthdata_status_fetch_fail():
    responses.add(
        responses.GET,
        "https://status.earthdata.nasa.gov/api/v1/statuses",
        body="Internal Server Error",
        status=500,
    )

    result = earthaccess.status()
    assert isinstance(result, dict)
    assert result == expected


def test_auth_returns_valid_auth_class():
    auth = earthaccess.login(strategy="environment")
    assert isinstance(auth, earthaccess.Auth)
    assert isinstance(earthaccess.__auth__, earthaccess.Auth)
    assert earthaccess.__auth__.authenticated


def test_dataset_search_returns_none_with_no_parameters():
    results = earthaccess.search_datasets()
    assert isinstance(results, list)
    assert len(results) == 0


@pytest.mark.parametrize("kwargs", dataset_valid_params)
def test_dataset_search_returns_valid_results(kwargs):
    results = earthaccess.search_datasets(**kwargs)
    assert isinstance(results, list)
    assert isinstance(results[0], dict)


def test_dataset_search_summary_missing_file_dist_info():
    results = earthaccess.search_datasets(short_name="AIRS_CPR_IND", daac="GES_DISC")
    collection_with_no_FileDistributionInformation = results[0]
    assert collection_with_no_FileDistributionInformation.summary()["file-type"] == ""


@pytest.mark.parametrize("kwargs", granules_valid_params)
def test_granules_search_returns_valid_results(kwargs):
    results = earthaccess.search_data(count=10, **kwargs)
    assert isinstance(results, list)
    assert len(results) <= 10


@pytest.mark.parametrize("selection", [0, slice(None)])
@pytest.mark.parametrize("use_url", [True, False])
def test_download(tmp_path, selection, use_url):
    results = earthaccess.search_data(
        count=2,
        short_name="ATL08",
        cloud_hosted=True,
        bounding_box=(-92.86, 16.26, -91.58, 16.97),
    )
    if use_url:
        # Download via file URL string instead of DataGranule object
        results = [link for r in results for link in r.data_links(access="indirect")]
    result = results[selection]
    files = earthaccess.download(result, str(tmp_path))
    assert isinstance(files, list)
    assert all(Path(f).exists() for f in files)


def fail_to_download_file(*args, **kwargs):
    raise IOError("Download failed")


def test_download_immediate_failure(tmp_path: Path):
    results = earthaccess.search_data(
        short_name="ATL06",
        bounding_box=(-10, 20, 10, 50),
        temporal=("1999-02", "2019-03"),
        count=3,
    )

    with patch.object(earthaccess.__store__, "_download_file", fail_to_download_file):
        with pytest.raises(IOError, match="Download failed"):
            # By default, we set pqdm exception_behavior to "immediate" so that
            # it simply propagates the first download error it encounters, halting
            # any further downloads.
            earthaccess.download(results, tmp_path, pqdm_kwargs=dict(disable=True))


def test_download_deferred_failure(tmp_path: Path):
    count = 3
    results = earthaccess.search_data(
        short_name="ATL06",
        bounding_box=(-10, 20, 10, 50),
        temporal=("1999-02", "2019-03"),
        count=count,
    )

    with patch.object(earthaccess.__store__, "_download_file", fail_to_download_file):
        # With "deferred" exceptions, pqdm catches all exceptions, then at the end
        # raises a single generic Exception, passing the sequence of caught exceptions
        # as arguments to the Exception constructor.
        with pytest.raises(Exception) as exc_info:
            earthaccess.download(
                results,
                tmp_path,
                pqdm_kwargs=dict(exception_behaviour="deferred", disable=True),
            )

    errors = exc_info.value.args
    assert len(errors) == count
    assert all(isinstance(e, IOError) and str(e) == "Download failed" for e in errors)


def test_auth_environ(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv("EARTHDATA_TOKEN", raising=False)
    earthaccess.login(strategy="environment")
    environ = earthaccess.auth_environ()
    assert environ == {
        "EARTHDATA_USERNAME": os.environ["EARTHDATA_USERNAME"],
        "EARTHDATA_PASSWORD": os.environ["EARTHDATA_PASSWORD"],
    }


def test_auth_environ_raises(monkeypatch):
    # Ensure `earthaccess.__auth__` always returns a new,
    # unauthenticated `earthaccess.Auth` instance, bypassing
    # automatic auth behavior
    monkeypatch.setattr(earthaccess, "__auth__", earthaccess.Auth())

    # Ensure `earthaccess.auth_environ()` raises an informative error
    # when not authenticated
    with pytest.raises(RuntimeError, match="authenticate"):
        earthaccess.auth_environ()



import logging

import earthaccess
import earthaccess.daac
import pytest
import requests
import s3fs
from earthaccess.exceptions import LoginStrategyUnavailable

logger = logging.getLogger(__name__)


def test_auth_can_read_earthdata_env_variables():
    auth = earthaccess.login(strategy="environment")
    logger.info(f"Current username: {auth.username}")
    logger.info(f"earthaccess version: {earthaccess.__version__}")

    assert isinstance(auth, earthaccess.Auth)
    assert isinstance(earthaccess.__auth__, earthaccess.Auth)
    assert earthaccess.__auth__.authenticated


def test_auth_can_read_from_netrc_file(mock_netrc):
    auth = earthaccess.login(strategy="netrc")
    assert auth.authenticated


def test_auth_strategy_unavailable_netrc_is_not_present(mock_missing_netrc):
    with pytest.raises(LoginStrategyUnavailable):
        earthaccess.login(strategy="netrc")


def test_auth_populates_attrs():
    auth = earthaccess.login(strategy="environment")
    assert isinstance(auth, earthaccess.Auth)
    assert isinstance(earthaccess.__auth__, earthaccess.Auth)
    assert earthaccess.__auth__.authenticated


def test_auth_can_create_authenticated_requests_sessions():
    session = earthaccess.get_requests_https_session()
    assert "Authorization" in session.headers
    assert "Bearer" in session.headers["Authorization"]  # type: ignore


@pytest.mark.parametrize(
    "daac", [daac for daac in earthaccess.daac.DAACS if daac["s3-credentials"]]
)
def test_auth_can_fetch_s3_credentials(daac):
    auth = earthaccess.login(strategy="environment")
    assert auth.authenticated

    try:
        credentials = earthaccess.get_s3_credentials(daac["short-name"])
    except requests.RequestException as e:
        logger.error(f"Failed to fetch S3 credentials: {e}")
    else:
        assert isinstance(credentials, dict)
        assert "accessKeyId" in credentials


@pytest.mark.parametrize("location", ({"daac": "podaac"}, {"provider": "pocloud"}))
def test_get_s3_credentials_lowercase_location(location):
    earthaccess.login(strategy="environment")
    creds = earthaccess.get_s3_credentials(**location)

    assert creds
    assert all(
        creds[key]
        for key in ["accessKeyId", "secretAccessKey", "sessionToken", "expiration"]
    )


@pytest.mark.parametrize("location", ({"daac": "podaac"}, {"provider": "pocloud"}))
def test_get_s3_filesystem_lowercase_location(location):
    earthaccess.login(strategy="environment")
    fs = earthaccess.get_s3_filesystem(**location)

    assert isinstance(fs, s3fs.S3FileSystem)
    assert all(fs.storage_options[key] for key in ["key", "secret", "token"])



"""Helper script that automatically skips the yamlfmt pre-commit hook on Windows.

This is not intended to be called directly.  It is configured to be called by
pre-commit via the configuration in .pre-commit-config.yml for the yamlfmt hook.

Due to https://github.com/google/yamlfmt/issues/263, comments in yaml files are
not handled properly under Windows, so any Windows user that causes yamlfmt to
format yaml files will cause CI build failure because we build under Linux, and
the resulting format will differ, causing file changes during the pre-commit
step, causing build failure.

Therefore, this script avoids calling yamlfmt (via pre-commit) when running on
Windows.
"""

import platform
import subprocess
import sys

if platform.system() != "Windows":
    sys.exit(subprocess.call(["yamlfmt", *sys.argv[1:]]))



"""Tests for search_data.  This needs a mock to be set up but for now, I get all
the first 10 granules using count=10.
"""

import earthaccess
import pytest

# Set SKIP_THIS to False to run test
SKIP_THIS = True

# Number of granules to return passed to count, also expected number returned
expected_count = 10

# I use ATL10 as a test case
short_name = "ATL10"
version = "006"  # This may need to be upped
provider = "NSIDC_CPRD"
daac = "NSIDC"
data_center = "NSIDC"
orbit_number = 436
granule_name = "ATL10-02_20181014000347_02350101_006_02.h5"

# Polygon over Jakobshavn Isbrae
polygon = [
    (-49.64860422604741, 69.23553485026147),
    (-49.667876114626296, 69.07309059285959),
    (-49.1722491331669, 69.03175841820749),
    (-47.53552489113113, 69.03872918462292),
    (-47.35616491854395, 69.22149993224824),
    (-48.1447695277283, 69.33507802083219),
    (-49.178671242118384, 69.29455117736225),
    (-49.64860422604741, 69.23553485026147),
]

# Canyon de Chelly, Az (Near Chinle, Naabeehó Bináhásdzo)
line = [
    (-109.49099495904196, 36.141620186146454),
    (-109.46326273316441, 36.12549319389707),
    (-109.42416829289516, 36.11336807764218),
    (-109.39942951589708, 36.10816134722084),
    (-109.33995221922424, 36.10943400693182),
    (-109.28600071040317, 36.11941578661717),
]

# Taos, NM
circle = (-105.61708725711999, 36.38510879364757, 1000.0)


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_shortname():
    """Tests granule search by short name returns
    results that match expected_count.
    """
    results = earthaccess.search_data(short_name=short_name, count=expected_count)
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_with_version():
    """Tests granule search by short name and version.
    returned expected_count.
    """
    results = earthaccess.search_data(
        short_name=short_name, version=version, count=expected_count
    )
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_provider():
    """Tests granule search by provider returns
    expected_count.
    """
    results = earthaccess.search_data(
        provider=provider,
        count=expected_count,
    )
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_daac():
    """Tests granule search by daac returns
    expected_count.
    """
    results = earthaccess.search_data(
        daac=daac,
        count=expected_count,
    )
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_data_center():
    """Tests granule search by data_center returns
    expected_count.
    """
    results = earthaccess.search_data(
        data_center=data_center,
        count=expected_count,
    )
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_coud_hosted():
    """Tests granule search by short_name with only
    cloud_hosted granules.
    """
    results = earthaccess.search_data(
        short_name=short_name,
        cloud_hosted=True,
        count=expected_count,
    )
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_online_only():
    """Tests granule search by short_name with online
    only granules.
    """
    results = earthaccess.search_data(
        short_name=short_name,
        online_only=True,
        count=expected_count,
    )
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_downloadable():
    """Tests granule search by short_name with downloadable
    granules.

    A granule is downloadable when it contains at least one RelatedURL of type GETDATA.
    """
    results = earthaccess.search_data(
        short_name=short_name,
        downloadable=True,
        count=expected_count,
    )
    assert len(results) == expected_count


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_orbit_number():
    """Tests granule search by short_name with orbit_number."""
    results = earthaccess.search_data(
        short_name=short_name,
        orbit_number=orbit_number,
        count=expected_count,
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_granule_name():
    """Tests granule search by short_name and granule name."""
    results = earthaccess.search_data(
        short_name=short_name,
        granule_name=granule_name,
    )
    assert len(results) == 1


# day_night_flag returns 0 granules
@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_day_night_flag():
    """Tests granule search by short_name and granule name."""
    results = earthaccess.search_data(
        short_name="MOD10A1", day_night_flag="Day", count=expected_count
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_daac_with_instrument():
    """Test search_data by daac and instrument."""
    results = earthaccess.search_data(
        daac="NSIDC",
        instrument="modis",
        count=expected_count,
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_daac_with_platform():
    """Test search_data by daac and platform."""
    results = earthaccess.search_data(
        daac="NSIDC",
        platform="terra",
        count=expected_count,
    )
    assert len(results) > 0


@pytest.mark.skip(reason="unable to build useful test")
def test_search_data_by_short_name_with_cloud_cover():
    """Test search_data by short_name and cloud cover."""
    results = earthaccess.search_data(
        short_name="MOD10A1",
        cloud_cover=(0, 50),
        count=10,
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_bounding_box():
    """Tests searching for granules with bounding box."""
    results = earthaccess.search_data(
        short_name="ATL06",
        bounding_box=(-46.5, 61.0, -42.5, 63.0),
        count=expected_count,
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_polygon():
    """Tests searching for granules with polygon."""
    results = earthaccess.search_data(
        short_name="ATL06",
        polygon=polygon,
        count=expected_count,
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_point():
    """Tests searching for granules with point."""
    results = earthaccess.search_data(
        short_name="MOD10A1",
        point=(-105.61708725711999, 36.38510879364757),
        count=10,
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_circle():
    """Tests searching for granules with circle."""
    results = earthaccess.search_data(
        short_name="ATL03",
        circle=circle,
        count=expected_count,
    )
    assert len(results) > 0


@pytest.mark.skipif(SKIP_THIS, reason="calls python-cmr, set SKIP_THIS=False to run")
def test_search_data_by_short_name_with_line():
    """Tests searching for granules with line."""
    results = earthaccess.search_data(
        short_name="ATL08",
        line=line,
        count=expected_count,
    )
    assert len(results) > 0



from earthaccess.formatters import (
    STATIC_FILES,
    _load_static_files,
    _repr_granule_html,
)
from earthaccess.results import DataGranule


def test_load_static_files():
    # We simply test that the number of static files loaded is the same as the
    # number of files in the STATIC_FILES list.  If we were to add logic to
    # check the contents of the files, then we would end up duplicating the
    # logic in the _load_static_files function, which wouldn't make sense to do.
    # If _load_static_files contains a bug, then this test will likely fail due
    # to the function raising an exception.
    assert len(_load_static_files()) == len(STATIC_FILES)


def test_repr_granule_html():
    static_contents = _load_static_files()
    size1 = 128573
    size2 = 2713600
    umm = {
        "RelatedUrls": [
            {
                "URL": "https://data.csdap.earthdata.nasa.gov/data.h5",
                "Type": "GET DATA",
            },
            {
                "URL": "s3://csda-cumulus-prod-protected-5047/data.h5",
                "Type": "GET DATA VIA DIRECT ACCESS",
            },
            {
                "URL": "https://data.csdap.earthdata.nasa.gov/thumb.jpg",
                "Type": "GET RELATED VISUALIZATION",
            },
        ],
        "DataGranule": {
            "ArchiveAndDistributionInformation": [
                {"SizeInBytes": size1},
                {"SizeInBytes": size2},
            ],
        },
    }

    html = _repr_granule_html(DataGranule({"umm": umm}, cloud_hosted=True))

    assert f"{round((size1 + size2) / 1024 / 1024, 2)} MB" in html
    assert [url["URL"] in html for url in umm["RelatedUrls"]] == [True, False, True]
    assert all(content in html for content in static_contents)



import logging
import unittest

import earthaccess
import pytest

logger = logging.getLogger(__name__)
assertions = unittest.TestCase("__init__")


logger.info(f"earthaccess version: {earthaccess.__version__}")


@pytest.fixture(
    scope="module",
    params=[
        ("MUR25-JPL-L4-GLOB-v04.2", 2),
        ("AVHRR_OI-NCEI-L4-GLOB-v2.1", 1),
        ("TEMPO_NO2_L3", 2),
        ("M2T1NXSLV", 1),
    ],
)
def granules(request):
    short_name, count = request.param
    granules = earthaccess.search_data(
        count=count, temporal=("2025"), short_name=short_name
    )
    return granules


def test_open_virtual_mfdataset(granules):
    # Simply check that the dmrpp can be found, parsed, and loaded. Actual parser result is checked in virtualizarr
    vds = earthaccess.open_virtual_mfdataset(granules, concat_dim="time")
    # We can use fancy indexing
    assert vds.isel(time=0) is not None



from typing import Any, List
from uuid import uuid4

import importlib_resources

STATIC_FILES = ["iso_bootstrap4.0.0min.css", "styles.css"]


def _load_static_files() -> List[str]:
    """Load styles."""
    return [
        importlib_resources.files("earthaccess.css").joinpath(fname).read_text("utf8")
        for fname in STATIC_FILES
    ]


def _repr_collection_html() -> str:
    return "<div></div>"


def _repr_granule_html(granule: Any) -> str:
    css_styles = _load_static_files()
    css_inline = f"""<div id="{uuid4()}" style="height: 0px; display: none">
            {"".join([f"<style>{style}</style>" for style in css_styles])}
            </div>"""
    style = "max-height: 120px;"
    dataviz_img = "".join(
        [
            f'<a href="{link}"><img style="{style}" src="{link}" alt="Data Preview"/></a>'
            for link in granule.dataviz_links()[:2]
            if link.startswith("http")
        ]
    )
    data_links = "".join(
        [
            f'<a href="{link}" target="_blank" class="btn btn-secondary btn-sm">{link.split("/")[-1]}</a>'
            for link in granule.data_links()
        ]
    )
    granule_size = round(granule.size(), 2)

    # TODO: probably this needs to be integrated on a list data structure
    return f"""
    {css_inline}
    <div class="bootstrap">
      <div class="container-fluid border">
        <div class="row border">
          <div class="col-6">
            <p><b>Data</b>: {data_links}<p/>
            <p><b>Size</b>: {granule_size} MB</p>
            <p><b>Cloud Hosted</b>: <span>{granule.cloud_hosted}</span></p>
          </div>
          <div class="col-2 offset-sm-3 pull-right">
            {dataviz_img}
          </div>
        </div>
      </div>
    </div>
    """



class LoginStrategyUnavailable(Exception):
    """The selected login strategy was skipped.

    This should be raised when a login strategy can't be attempted, for example because
    "environment" was selected but the envvars are not populated.

    DO NOT raise this exception when a login strategy is attempted and fails. For
    example, this exception would not be thrown when credentials were rejected;
    a `LoginAttemptFailure` should be thrown instead.
    """

    pass


class LoginAttemptFailure(Exception):
    """The login attempt failed.

    This should be raised when a login attempt fails, for example, because
    the user's credentials were rejected.

    DO NOT raise this exception when a login strategy can't be attempted. For
    example, this exception would not be thrown when "environment" was selected
    but the envvars are not populated; a `LoginStrategyUnavailable` should be
    thrown instead.
    """

    pass


class DownloadFailure(Exception):
    """The download attempt failed.

    This should be raised when a download attempt fails, for example, because
    the file could not be retrieved or the download process was interrupted.
    """

    pass


class ServiceOutage(Exception):
    """A service outage has been detected.

    This should be raised when Earthdata services are unavailable or experiencing
    outages that prevent normal operations.
    """

    pass


class EulaNotAccepted(DownloadFailure):
    """The user has not accepted the EULA.

    This should be raised when a user attempts to access data that requires
    EULA acceptance, but they have not accepted the EULA.
    """

    pass
